import dataclasses
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import (
    TypeVar,
    Set,
    Type,
    Dict,
    Optional,
    Iterable,
    MutableMapping,
    Union,
    Tuple,
    List,
    Callable,
    Generic,
    Any,
    cast,
    overload,
)
from weakref import WeakValueDictionary


from ofrak.model.tag_model import ResourceTag
from ofrak_type.range import Range

T = TypeVar("T")
RT = TypeVar("RT", bound="ResourceTag")
RA = TypeVar("RA", bound="ResourceAttributes")
X = TypeVar("X", str, int, float, bytes)  # Indexable field types

_INDEXABLE_TYPES: Dict[str, Type] = {
    indexable_type.__name__: indexable_type
    for indexable_type in getattr(X, "__constraints__")  # type: ignore
}


class ResourceIndexedAttribute(Generic[X]):
    """
    Descriptor class for values in resource attributes which can be indexed. When a field `Foo`
    of a [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes] type `A` is indexed,
    it is possible to include an `r_filter` or `r_sort` in a query to the resource service which
    filters the returned resource by the value of `foo` each of them have.

    This class should not be explicitly instantiated, instead created using the @index decorator.

    For example:
    ```python
    @dataclass
    class A(ResourceAttributes):
        x: int

        @index
        def Foo(self) -> int:
            return self.x
    ```
    """

    def __init__(
        self,
        getter_func: Callable[[Any], X],
        uses_indexes: Iterable["ResourceIndexedAttribute"] = (),
    ):
        """
        :param getter_func: Getter function for the property
        :param uses_indexes: Additional index types that are required to calculate the value
        of this index

        :raises TypeError: if the getter function does not have a return type annotation
        :raises TypeError: if the getter does not return an indexable type
        """
        _validate_indexed_type(getter_func)
        self.fget: Callable[[Any], X] = getter_func
        self.attributes_owner: Optional[Type[ResourceAttributes]] = None
        self.uses_indexes = uses_indexes
        self.used_by_indexes: List["ResourceIndexedAttribute"] = []
        self.index_name: str = getter_func.__name__

        for other_index in self.uses_indexes:
            other_index.used_by_indexes.append(self)

    def __set_name__(self, owner, name):
        self.attributes_owner = owner
        self.__name__ = f"{owner.__name__}.{name}"
        self.index_name = name

    @overload
    def __get__(self, instance: None, owner: type) -> "ResourceIndexedAttribute[X]":
        """
        Applicable when getting the ResourceIndexedAttribute of a class, not an instance.
        Example (continued building off of example from class docstring):
        A.X  # Returns a ResourceIndexedAttribute[int]
        """
        ...

    @overload
    def __get__(self, instance: Any, owner: type) -> X:
        """
        Applicable when getting the ResourceIndexedAttribute of an instance
        Example (continued building off of example from class docstring):
        a = A(10)
        a.X  # Returns 10
        """
        ...

    def __get__(
        self, instance: Optional[Any], owner: type
    ) -> Union["ResourceIndexedAttribute[X]", X]:
        if instance is None:
            return self
        else:
            return self.fget(instance)

    def __getattr__(self, name) -> Any:
        ...

    def __set__(self, instance, value):
        raise ValueError("Cannot set value of indexed attributes")

    def get_value(
        self,
        index_holder: "ResourceModel",
    ) -> Optional[X]:
        if self.attributes_owner is None:
            raise TypeError(
                f"Cannot get index value for {self.__name__} of model "
                f"{index_holder.id.hex()} because {self.__name__}'s owner has not "
                f"been set. This cannot happen unless `get_index` has somehow been "
                f"called during class creation, before the owner is set."
            )
        else:
            attributes = index_holder.get_attributes(self.attributes_owner)
            if attributes is None:
                return None
            elif self.uses_indexes:
                # Create new copy of attributes to inject index values into
                attributes_plus_required_indexes = dataclasses.replace(attributes)
                for nested_index in self.uses_indexes:
                    val = nested_index.get_value(index_holder)
                    if val is None:
                        # Not all of the nested indexes are available, can't calculate index val.
                        return None
                    object.__setattr__(
                        attributes_plus_required_indexes, nested_index.index_name, val
                    )
                return self.fget(attributes_plus_required_indexes)
            else:
                return self.fget(attributes)

    def __repr__(self) -> str:
        return self.__name__


@overload
def index(
    *,
    uses_indexes: Iterable[ResourceIndexedAttribute] = ...,
) -> Callable[[Callable[[Any], X]], ResourceIndexedAttribute[X]]:
    """
    When called as:

    @index(uses_indexes=(...))
    def MyIndex(self):
        ...
    """
    ...


@overload
def index(
    index_value_getter: Callable[[Any], X],
) -> ResourceIndexedAttribute[X]:
    """
    When called as:

    @index
    def MyIndex(self):
        ...
    """
    ...


def index(
    index_value_getter: Callable[[Any], X] = None,
    *,
    uses_indexes: Iterable[ResourceIndexedAttribute] = (),
) -> Union[
    Callable[[Callable[[Any], X]], ResourceIndexedAttribute[X]], ResourceIndexedAttribute[X]
]:
    """
    Create a new indexable attribute for a
    [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes].

    :param index_value_getter: Method of
        [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes] which returns the
        value of the index for that instance.
    :param uses_indexes: Additional index types that are required to calculate the value
    of this index.

    :return: [ResourceIndexedAttribute][ofrak.model.resource_model.ResourceIndexedAttribute]
        instance
    """
    # See if we're being called as @index or @index().
    if index_value_getter is None:
        # We're called with parens.
        def wrap(_index_value_getter) -> ResourceIndexedAttribute[X]:
            return ResourceIndexedAttribute[X](_index_value_getter, uses_indexes)

        return wrap  # type: ignore

    # We're called as @index without parens.
    return ResourceIndexedAttribute[X](index_value_getter)


class ResourceAttributes:
    DATACLASS_PARAMS = {"frozen": True, "eq": True}

    """
    Wraps immutable attributes attached to a resource. While not enforced programmatically, only
    analyzers should add/replace attributes attached to a resource. Additionally, a
    [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes] instance also defines
    which component attached the attributes to a specific resource.
    """

    def __str__(self):
        fields_str = ", ".join(
            f"{field.name}={str(getattr(self, field.name))}" for field in dataclasses.fields(self)
        )
        return f"{self.__class__.__name__}({fields_str})"

    @classmethod
    def get_indexable_attributes(cls) -> List[ResourceIndexedAttribute]:
        indexable_attributes = []
        for name, attr in cls.__dict__.items():
            if type(attr) is ResourceIndexedAttribute:
                indexable_attributes.append(attr)
        return indexable_attributes

    @staticmethod
    def replace_updated(resource_attributes: RA, updated_attributes: Any) -> RA:
        """
        Replace the fields of `resource_attributes` with the updated values found in
        `updated_attributes`, returning a new object. The fields having non-`None` values in
        `updated_attributes` are considered to be updated and will be replaced in
        `resource_attributes` if they exist there.

        Both arguments must be `dataclass` instances.

        `updated_attributes` is typically a descendant of
        [ComponentConfig][ofrak.model.component_model.ComponentConfig].

        !!! todo "To do"

               This currently assumes that values can't be updated to `None`, but that could happen.

        :raises TypeError: if any of `resource_attributes` or `updated_attributes` isn't a
            dataclass instance.
        """
        for obj in (resource_attributes, updated_attributes):
            if not (dataclasses.is_dataclass(obj) and not isinstance(obj, type)):
                raise TypeError(f"{obj.__name__} must be a dataclass instance")
        updated_fields = {
            field: val
            for field, val in dataclasses.asdict(updated_attributes).items()
            if val is not None
        }
        updated_attributes = dataclasses.replace(
            resource_attributes,
            **updated_fields,
        )
        return updated_attributes


class ResourceAttributeDependency:
    dependent_resource_id: bytes
    component_id: bytes
    attributes: Type[ResourceAttributes]

    __slots__ = ("dependent_resource_id", "component_id", "attributes")

    def __init__(
        self,
        dependent_resource_id: bytes,
        component_id: bytes,
        attributes: Type[ResourceAttributes],
    ):
        """
        Create a dependency that says the resource ``dependent_resource_id`` has some attributes
        ``attributes`` (added by ``component_id``) which depend on some information in another
        resource. That information is either a range of the data or some attributes of that other
        resource.

        :param dependent_resource_id: ID of the resource which has a dependency
        :param component_id: Type of attributes on the resource which has a dependency
        :param attributes: Component which ran on the resource to create the attributes
        """
        self.dependent_resource_id = dependent_resource_id
        self.component_id = component_id
        self.attributes = attributes

    def __hash__(self):
        return hash((self.dependent_resource_id, self.component_id, self.attributes))

    def __eq__(self, other):
        if not isinstance(other, ResourceAttributeDependency):
            return False
        return (
            self.dependent_resource_id == other.dependent_resource_id
            and self.component_id == other.component_id
            and self.attributes == other.attributes
        )


ModelIdType = bytes
ModelDataIdType = Optional[bytes]
ModelParentIdType = Optional[bytes]
ModelTagsType = Set[ResourceTag]
ModelAttributesType = Dict[Type[RA], RA]
ModelDataDependenciesType = Dict[ResourceAttributeDependency, Set[Range]]
ModelAttributeDependenciesType = Dict[Type[ResourceAttributes], Set[ResourceAttributeDependency]]
ModelComponentVersionsType = Dict[bytes, int]
ModelComponentsByAttributesType = Dict[Type[ResourceAttributes], Tuple[bytes, int]]


class ResourceModel:
    """
    :param ModelIdType id:
    :param ModelDataIdType data_id:
    :param ModelParentIdType parent_id:
    :param Optional[Set[ResourceTag]] tags:
    :param Optional[ModelAttributesType] attributes:
    :param Optional[ModelAttributesType] data_dependencies: Stores the dependencies of other
        resources on specific data ranges within this resource
    :param Optional[ModelAttributeDependenciesType] attribute_dependencies: Stores the dependencies
        of other resources on [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes]
        of this resource
    :param Optional[ModelComponentVersionsType] component_versions: Stores the version of
        [ComponentInterface][ofrak.component.interface.ComponentInterface] which has been run on
        this resource
    :param Optional[ModelComponentsByAttributesType] components_by_attributes: For each
        [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes], stores the id of the
        [ComponentInterface][ofrak.component.interface.ComponentInterface] which was run to
        create it
    """

    id: ModelIdType
    data_id: ModelDataIdType
    parent_id: ModelParentIdType
    tags: ModelTagsType
    attributes: ModelAttributesType
    data_dependencies: ModelDataDependenciesType
    attribute_dependencies: ModelAttributeDependenciesType
    component_versions: ModelComponentVersionsType
    components_by_attributes: ModelComponentsByAttributesType

    __slots__ = (
        "id",
        "data_id",
        "parent_id",
        "tags",
        "attributes",
        "data_dependencies",
        "attribute_dependencies",
        "component_versions",
        "components_by_attributes",
        "__weakref__",  # Required to get a weak ref to an instance
    )

    def __init__(
        self,
        id: ModelIdType,
        data_id: ModelDataIdType = None,
        parent_id: ModelParentIdType = None,
        tags: Optional[Set[ResourceTag]] = None,
        attributes: Optional[ModelAttributesType] = None,
        data_dependencies: Optional[ModelDataDependenciesType] = None,
        attribute_dependencies: Optional[ModelAttributeDependenciesType] = None,
        component_versions: Optional[ModelComponentVersionsType] = None,
        components_by_attributes: Optional[ModelComponentsByAttributesType] = None,
    ):
        attributes = attributes or {}
        data_dependencies = defaultdict(set, data_dependencies or {})
        attribute_dependencies = defaultdict(set, attribute_dependencies or {})
        component_versions = component_versions or {}
        components_by_attributes = components_by_attributes or {}

        self.id: ModelIdType = id
        self.data_id: ModelDataIdType = data_id
        self.parent_id: ModelParentIdType = parent_id
        self.tags: ModelTagsType = set(tags) if tags else set()
        self.attributes: ModelAttributesType = attributes
        self.data_dependencies: ModelDataDependenciesType = data_dependencies
        self.attribute_dependencies: ModelAttributeDependenciesType = attribute_dependencies
        self.component_versions: ModelComponentVersionsType = component_versions
        self.components_by_attributes: ModelComponentsByAttributesType = components_by_attributes

    @staticmethod
    def _clone_dependencies(
        dependencies: ModelAttributeDependenciesType,
    ) -> ModelAttributeDependenciesType:
        new_dependencies: ModelAttributeDependenciesType = defaultdict(set)
        new_dependencies.update(
            {attribute: set(dependencies) for attribute, dependencies in dependencies.items()}
        )
        return new_dependencies

    @staticmethod
    def _clone_data_dependencies(
        dependencies: Dict[ResourceAttributeDependency, Set[Range]]
    ) -> Dict[ResourceAttributeDependency, Set[Range]]:
        new_dependencies = defaultdict(set)
        for dependency, ranges in dependencies.items():
            new_dependencies[dependency] = set(ranges)
        return new_dependencies

    def get_tags(self, inherit: bool = True) -> Set[ResourceTag]:
        if inherit is False:
            return set(self.tags)
        tags = set()
        for _tag in self.tags:
            tags.update(_tag.tag_classes())
        return tags

    def get_most_specific_tags(self) -> Iterable[ResourceTag]:
        most_specific_tags = set(self.tags)
        for tag in [
            base for tag in self.tags for base in cast(Iterable[ResourceTag], tag.__bases__)
        ]:
            most_specific_tags.discard(tag)
        return most_specific_tags

    @property
    def caption(self) -> str:
        captions = []
        for tag in self.get_most_specific_tags():
            captions.append(tag.caption(self.attributes))
        return ", ".join(captions)

    def has_tag(self, tag: ResourceTag, inherit: bool = True) -> bool:
        has_tag = tag in self.tags
        if has_tag or inherit is False:
            return has_tag
        for _tag in self.tags:
            if tag in _tag.tag_classes():
                return True
        return False

    def get_attributes(self, attributes_type: Type[RA]) -> Optional[RA]:
        return self.attributes.get(attributes_type, None)

    def get_index_values(self) -> Dict[ResourceIndexedAttribute[X], X]:
        indexable_values = dict()
        for attributes in self.attributes.values():
            for indexable_attribute in attributes.get_indexable_attributes():
                indexable_values[indexable_attribute] = indexable_attribute.get_value(self)
        return indexable_values

    def has_attributes(self, attributes_type: Type[ResourceAttributes]) -> bool:
        return attributes_type in self.attributes

    def get_component_id_by_attributes(
        self, attributes: Type[ResourceAttributes]
    ) -> Optional[bytes]:
        component = self.components_by_attributes.get(attributes)
        if component is not None:
            component_id, _ = component
            return component_id
        return None

    def get_component_version(self, component_id: bytes) -> Optional[int]:
        return self.component_versions.get(component_id)

    def clone(self):
        return ResourceModel(
            self.id,
            self.data_id,
            self.parent_id,
            set(self.tags),
            dict(self.attributes),
            ResourceModel._clone_data_dependencies(self.data_dependencies),
            ResourceModel._clone_dependencies(self.attribute_dependencies),
            dict(self.component_versions),
            dict(self.components_by_attributes),
        )

    @staticmethod
    def create(
        id: bytes,
        data_id: Optional[bytes] = None,
        parent_id: Optional[bytes] = None,
        tags: Iterable[ResourceTag] = None,
        attributes: Iterable[ResourceAttributes] = None,
        created_by_component_id: bytes = None,
        created_by_component_version: int = None,
    ) -> "ResourceModel":
        attributes_dict = dict()
        components_by_attributes = dict()
        if attributes is not None:
            for _attributes in attributes:
                attributes_dict[type(_attributes)] = _attributes
                if created_by_component_id is not None and created_by_component_version is not None:
                    components_by_attributes[type(_attributes)] = (
                        created_by_component_id,
                        created_by_component_version,
                    )
        components_by_version: ModelComponentVersionsType = dict()

        final_tags: Set[ResourceTag] = set()

        if tags is not None:
            final_tags.update(tags)
            _tags_to_process = list(tags)

            while len(_tags_to_process) > 0:
                t = _tags_to_process.pop()
                for base_tag in t.base_tags():
                    if base_tag not in final_tags:
                        final_tags.add(base_tag)
                        _tags_to_process.append(base_tag)
        return ResourceModel(
            id,
            data_id,
            parent_id,
            final_tags,
            attributes_dict,
            components_by_attributes=components_by_attributes,
            component_versions=components_by_version,
        )

    def __eq__(self, other):
        if not isinstance(other, ResourceModel):
            return False

        return all(
            [
                self.id == other.id,
                self.data_id == other.data_id,
                self.parent_id == other.parent_id,
                self.tags == other.tags,
                self.attributes == other.attributes,
                self.data_dependencies == other.data_dependencies,
                self.attribute_dependencies == other.attribute_dependencies,
                self.component_versions == other.component_versions,
                self.components_by_attributes == other.components_by_attributes,
            ]
        )


# NOTE: in Python >= 3.10, @dataclass(slots=True) can be used to make this class definition smaller.
class ResourceModelDiff:
    id: bytes
    tags_added: Set[ResourceTag]
    tags_removed: Set[ResourceTag]
    attributes_added: Dict[Type[ResourceAttributes], ResourceAttributes]
    attributes_removed: Set[Type[ResourceAttributes]]
    data_dependencies_added: Set[Tuple[ResourceAttributeDependency, Range]]
    data_dependencies_removed: Set[ResourceAttributeDependency]
    attribute_dependencies_added: Set[Tuple[Type[ResourceAttributes], ResourceAttributeDependency]]
    attribute_dependencies_removed: Set[
        Tuple[Type[ResourceAttributes], ResourceAttributeDependency]
    ]
    component_versions_added: Set[Tuple[bytes, int]]
    component_versions_removed: Set[bytes]
    attributes_component_added: Set[Tuple[Type[ResourceAttributes], bytes, int]]
    attributes_component_removed: Set[Type[ResourceAttributes]]

    __slots__ = (
        "id",
        "tags_added",
        "tags_removed",
        "attributes_added",
        "attributes_removed",
        "data_dependencies_added",
        "data_dependencies_removed",
        "attribute_dependencies_added",
        "attribute_dependencies_removed",
        "component_versions_added",
        "component_versions_removed",
        "attributes_component_added",
        "attributes_component_removed",
    )

    def __init__(
        self,
        id: bytes,
        tags_added: Optional[Set[ResourceTag]] = None,
        tags_removed: Optional[Set[ResourceTag]] = None,
        attributes_added: Optional[Dict[Type[ResourceAttributes], ResourceAttributes]] = None,
        attributes_removed: Optional[Set[Type[ResourceAttributes]]] = None,
        data_dependencies_added: Optional[Set[Tuple[ResourceAttributeDependency, Range]]] = None,
        data_dependencies_removed: Optional[Set[ResourceAttributeDependency]] = None,
        attribute_dependencies_added: Optional[
            Set[Tuple[Type[ResourceAttributes], ResourceAttributeDependency]]
        ] = None,
        attribute_dependencies_removed: Optional[
            Set[Tuple[Type[ResourceAttributes], ResourceAttributeDependency]]
        ] = None,
        component_versions_added: Optional[Set[Tuple[bytes, int]]] = None,
        component_versions_removed: Optional[Set[bytes]] = None,
        attributes_component_added: Optional[
            Set[Tuple[Type[ResourceAttributes], bytes, int]]
        ] = None,
        attributes_component_removed: Optional[Set[Type[ResourceAttributes]]] = None,
    ):
        self.id = id
        self.tags_added: Set[ResourceTag] = tags_added or set()
        self.tags_removed: Set[ResourceTag] = tags_removed or set()
        self.attributes_added: Dict[Type[ResourceAttributes], ResourceAttributes] = (
            attributes_added or {}
        )
        self.attributes_removed: Set[Type[ResourceAttributes]] = attributes_removed or set()
        self.data_dependencies_added: Set[Tuple[ResourceAttributeDependency, Range]] = (
            data_dependencies_added or set()
        )
        self.data_dependencies_removed: Set[ResourceAttributeDependency] = (
            data_dependencies_removed or set()
        )
        self.attribute_dependencies_added: Set[
            Tuple[Type[ResourceAttributes], ResourceAttributeDependency]
        ] = (attribute_dependencies_added or set())
        self.attribute_dependencies_removed = attribute_dependencies_removed or set()
        self.component_versions_added = component_versions_added or set()
        self.component_versions_removed = component_versions_removed or set()
        self.attributes_component_added = attributes_component_added or set()
        self.attributes_component_removed = attributes_component_removed or set()

    def apply(self, model: ResourceModel) -> ResourceModel:
        # TODO: Raise exceptions if trying to add stuff that already exist or remove stuff that
        #  doesn't
        updated_model = model.clone()
        updated_model.tags.update(self.tags_added)
        updated_model.tags.difference_update(self.tags_removed)

        for attributes_type_removed in self.attributes_removed:
            del updated_model.attributes[attributes_type_removed]
        for type_attributes_added, attributes_added in self.attributes_added.items():
            updated_model.attributes[type_attributes_added] = attributes_added

        for data_dependency in self.data_dependencies_removed:
            del updated_model.data_dependencies[data_dependency]
        for data_dependency, data_range in self.data_dependencies_added:
            updated_model.data_dependencies[data_dependency].add(data_range)

        for attribute_type, attribute_dependency in self.attribute_dependencies_removed:
            if attribute_dependency in updated_model.attribute_dependencies[attribute_type]:
                updated_model.attribute_dependencies[attribute_type].remove(attribute_dependency)
        for attribute_type, attribute_dependency in self.attribute_dependencies_added:
            updated_model.attribute_dependencies[attribute_type].add(attribute_dependency)

        for component_id in self.component_versions_removed:
            del updated_model.component_versions[component_id]
        for component_id, component_version in self.component_versions_added:
            updated_model.component_versions[component_id] = component_version

        for attribute_type in self.attributes_component_removed:
            if updated_model.get_component_id_by_attributes(attribute_type) is not None:
                del updated_model.components_by_attributes[attribute_type]
        for attribute_type, component_id, version in self.attributes_component_added:
            updated_model.components_by_attributes[attribute_type] = (component_id, version)

        return updated_model


class MutableResourceModel(ResourceModel):
    __slots__ = "is_modified", "diff", "is_deleted"

    def __init__(
        self,
        id: bytes,
        data_id: Optional[bytes] = None,
        parent_id: Optional[bytes] = None,
        tags: Optional[Set[ResourceTag]] = None,
        attributes: Optional[Dict[Type[ResourceAttributes], ResourceAttributes]] = None,
        data_dependencies: Optional[Dict[ResourceAttributeDependency, Set[Range]]] = None,
        attribute_dependencies: Optional[
            Dict[Type[ResourceAttributes], Set[ResourceAttributeDependency]]
        ] = None,
        component_versions: Optional[Dict[bytes, int]] = None,
        components_by_attributes: Optional[ModelComponentsByAttributesType] = None,
    ):
        super().__init__(
            id,
            data_id,
            parent_id,
            tags,
            attributes,
            data_dependencies,
            attribute_dependencies,
            component_versions,
            components_by_attributes,
        )
        self.is_modified = False
        self.is_deleted = False
        self.diff = ResourceModelDiff(self.id)

    @staticmethod
    def from_model(model: ResourceModel):
        return MutableResourceModel(
            model.id,
            model.data_id,
            model.parent_id,
            model.tags,
            dict(model.attributes),
            ResourceModel._clone_data_dependencies(model.data_dependencies),
            ResourceModel._clone_dependencies(model.attribute_dependencies),
            dict(model.component_versions),
            dict(model.components_by_attributes),
        )

    def add_tag(self, tag: ResourceTag) -> Set[ResourceTag]:
        if tag in self.tags:
            return set()
        self.is_modified = True
        self.tags.add(tag)
        self.diff.tags_added.add(tag)
        new_tags = {tag}

        for base_tag in tag.base_tags():
            new_tags.update(self.add_tag(base_tag))

        return new_tags

    def remove_tag(self, tag: ResourceTag):
        self.is_modified = True
        self.tags.remove(tag)
        self.diff.tags_removed.add(tag)

    def add_attributes(self, attributes: ResourceAttributes):
        attributes_type = type(attributes)
        prev_attributes = self.attributes.get(attributes_type)
        if prev_attributes is not None:
            if prev_attributes == attributes:
                return
            else:
                self.diff.attributes_removed.add(attributes_type)
        self.is_modified = True
        self.attributes[attributes_type] = attributes
        self.diff.attributes_added[attributes_type] = attributes

    def remove_attributes(self, attributes_type: Type[ResourceAttributes]):
        self.is_modified = True
        del self.attributes[attributes_type]
        self.diff.attributes_removed.add(attributes_type)

    def add_attribute_dependency(
        self, attribute_type: Type[ResourceAttributes], dependency: ResourceAttributeDependency
    ):
        self.is_modified = True
        self.attribute_dependencies[attribute_type].add(dependency)
        self.diff.attribute_dependencies_added.add((attribute_type, dependency))

    def add_data_dependency(self, dependency: ResourceAttributeDependency, data_range: Range):
        self.is_modified = True
        self.data_dependencies[dependency].add(data_range)
        self.diff.data_dependencies_added.add((dependency, data_range))

    def remove_dependency(self, dependency: ResourceAttributeDependency):
        if dependency in self.data_dependencies:
            self.is_modified = True
            del self.data_dependencies[dependency]
            self.diff.data_dependencies_removed.add(dependency)
        for attribute_type, attribute_dependencies in self.attribute_dependencies.items():
            if dependency in attribute_dependencies:
                self.is_modified = True
                attribute_dependencies.remove(dependency)
                self.diff.attribute_dependencies_removed.add((attribute_type, dependency))

    def add_component(
        self,
        component_id: bytes,
        version: int,
    ):
        """
        Mark that a component has run on a resource.
        """
        self.is_modified = True
        self.component_versions[component_id] = version
        self.diff.component_versions_added.add((component_id, version))

    def add_component_for_attributes(
        self,
        component_id: bytes,
        version: int,
        attribute_type: Type[ResourceAttributes],
    ):
        """
        Mark that a component has added attributes to a resource
        """
        self.components_by_attributes[attribute_type] = (component_id, version)
        self.diff.attributes_component_added.add((attribute_type, component_id, version))

    def remove_component(
        self, component_id: bytes, attribute_type: Optional[Type[ResourceAttributes]]
    ):
        if component_id in self.component_versions:
            self.is_modified = True
            del self.component_versions[component_id]
            self.diff.component_versions_removed.add(component_id)
        if attribute_type in self.components_by_attributes:
            self.is_modified = True
            del self.components_by_attributes[attribute_type]
            self.diff.attributes_component_removed.add(attribute_type)

    def reset(self, model: ResourceModel):
        self.id = model.id
        self.data_id = model.data_id
        self.parent_id = model.parent_id
        self.tags = model.tags
        self.attributes = dict(model.attributes)
        self.data_dependencies: Dict[
            ResourceAttributeDependency, Set[Range]
        ] = ResourceModel._clone_data_dependencies(model.data_dependencies)
        self.attribute_dependencies: ModelAttributeDependenciesType = (
            ResourceModel._clone_dependencies(model.attribute_dependencies)
        )
        self.component_versions = dict(model.component_versions)
        self.components_by_attributes = dict(model.components_by_attributes)
        self.is_modified = False
        self.diff = ResourceModelDiff(self.id)

    def save(self):
        self.is_modified = False
        diff = self.diff
        self.diff = ResourceModelDiff(self.id)
        return diff


class ResourceContext(ABC):
    """
    Resource context.
    """

    def __init__(self, resource_models: MutableMapping[bytes, MutableResourceModel]):
        self.resource_models = resource_models


class ResourceContextFactory(ABC):
    @abstractmethod
    def create(self) -> ResourceContext:
        pass


class EphemeralResourceContext(ResourceContext):
    def __init__(self):
        super().__init__(dict())


class EphemeralResourceContextFactory(ResourceContextFactory):
    def create(self) -> EphemeralResourceContext:
        return EphemeralResourceContext()


class ClientResourceContext(ResourceContext):
    def __init__(self):
        super().__init__(WeakValueDictionary())


class ClientResourceContextFactory(ResourceContextFactory):
    def create(self) -> ClientResourceContext:
        return ClientResourceContext()


def _validate_indexed_type(getter_func: Callable[[Any], X]):
    """
    Verify the getter function returns a valid indexable type - a primitive type which can be
    compared.

    :param getter_func:

    :raises TypeError: if the getter function does not have a return type annotation
    :raises TypeError: if the getter does not return an indexable type
    :return:
    """

    if not hasattr(getter_func, "__annotations__"):
        raise TypeError(
            f"Index {getter_func.__name__} must have type annotations, including return type"
        )
    annotations = getattr(getter_func, "__annotations__")
    if "return" not in annotations:
        raise TypeError(
            f"Index {getter_func.__name__} must have type annotations, including return type"
        )
    index_type = annotations["return"]
    if type(index_type) is str:
        # handles case where type annotations is "stringified" and not the actual type, e.g.
        # def foo(self) -> "int": ...
        index_type_name = index_type
    else:
        index_type_name = index_type.__name__

    if index_type_name not in _INDEXABLE_TYPES:
        raise TypeError(
            f"Type of index {getter_func.__name__} is {index_type}, which is not "
            f"one of {_INDEXABLE_TYPES.values()}; cannot index by this value!"
        )


@dataclasses.dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Data(ResourceAttributes):
    """
    Special attributes class for accessing info about a resource's binary data.
    Users should never access or modify this directly! Changing the fields of this data structure
    will not change any about the resource's data! At best, it will do nothing, and at worst it
    will screw up sorting/filtering.
    """

    _offset: int
    _length: int

    @index
    def Offset(self) -> int:
        return self._offset

    @index
    def Length(self) -> int:
        return self._length



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Tuple, Optional

from ofrak.component.analyzer import Analyzer
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.resource import Resource
from ofrak.resource_view import ResourceView
from pytest_ofrak.mock_component_types import MockAnalyzer, MockUnpacker


##################################################################################
# A mock library of ResourceAttributes, Resources, and Analyzers for tests to use
##################################################################################


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class AbstractionAttributesA(ResourceAttributes):
    pass


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class AbstractionAttributesB(ResourceAttributes):
    pass


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class AbstractionAttributesC(ResourceAttributes):
    pass


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class AbstractionAttributesD(ResourceAttributes):
    pass


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class AbstractionAttributesUnknown(ResourceAttributes):
    pass


class AbstractionCommon(ResourceView):
    pass


class AbstractionP(AbstractionCommon):
    pass


class AbstractionQ(AbstractionCommon):
    pass


class AbstractionR(AbstractionCommon):
    pass


class AbstractionRR(AbstractionR):
    pass


class ITargetsRROutputsD(Analyzer[None, Tuple[AbstractionAttributesD]], ABC):
    targets = (AbstractionRR,)
    outputs = (AbstractionAttributesD,)
    id = b"TargetsRROutputsD"

    @abstractmethod
    async def analyze(
        self, resource: Resource, config: Optional[ComponentConfig]
    ) -> Tuple[AbstractionAttributesD]:
        raise NotImplementedError()


class TargetsRROutputsD(MockAnalyzer, ITargetsRROutputsD):
    pass


class ITargetsPOutputsA(Analyzer[None, Tuple[AbstractionAttributesA]], ABC):
    targets = (AbstractionP,)
    outputs = (AbstractionAttributesA,)
    id = b"TargetsPOutputsA"

    @abstractmethod
    async def analyze(
        self, resource: Resource, config: Optional[ComponentConfig]
    ) -> Tuple[AbstractionAttributesA]:
        raise NotImplementedError()


class TargetsPOutputsA(MockAnalyzer, ITargetsPOutputsA):
    pass


class ITargetsCommonOutputsA(Analyzer[None, Tuple[AbstractionAttributesA]], ABC):
    targets = (AbstractionCommon,)
    outputs = (AbstractionAttributesA,)
    id = b"TargetsCommonOutputsA"

    @abstractmethod
    def analyze(
        self, resource: Resource, config: Optional[ComponentConfig]
    ) -> Tuple[AbstractionAttributesA]:
        raise NotImplementedError()


class TargetsCommonOutputsA(MockAnalyzer, ITargetsCommonOutputsA):
    pass


class ITargetsQOutputsABC(
    Analyzer[None, Tuple[AbstractionAttributesA, AbstractionAttributesB, AbstractionAttributesC]],
    ABC,
):
    targets = (AbstractionQ,)
    outputs = (AbstractionAttributesA, AbstractionAttributesB, AbstractionAttributesC)
    id = b"TargetsQOutputsABC"

    @abstractmethod
    def analyze(
        self, resource: Resource, config: Optional[ComponentConfig]
    ) -> Tuple[AbstractionAttributesA, AbstractionAttributesB, AbstractionAttributesC]:
        raise NotImplementedError()


class TargetsQOutputsABC(MockAnalyzer, ITargetsQOutputsABC):
    pass


class IWithoutImplementation(Analyzer[None, Tuple[AbstractionAttributesA]], ABC):
    targets = (AbstractionQ,)
    outputs = (AbstractionAttributesA,)
    id = b"WithoutImplementation"

    @abstractmethod
    def analyze(
        self, resource: Resource, config: Optional[ComponentConfig]
    ) -> Tuple[AbstractionAttributesA]:
        raise NotImplementedError()


class AbstractionPUnpacker(MockUnpacker):
    targets = [AbstractionP]


class AbstractionRUnpacker(MockUnpacker):
    targets = [AbstractionR]


class AbstractionRRUnpacker(MockUnpacker):
    targets = [AbstractionRR]



from abc import ABCMeta, abstractmethod
from typing import Optional, Tuple

from ofrak.model.component_model import ComponentRunResult
from ofrak.model.job_model import (
    JobModel,
    JobRunContext,
)
from ofrak.model.job_request_model import (
    JobAnalyzerRequest,
    JobComponentRequest,
    JobMultiComponentRequest,
)
from ofrak.service.abstract_ofrak_service import AbstractOfrakService
from ofrak.service.component_locator_i import ComponentFilter


class JobServiceInterface(AbstractOfrakService, metaclass=ABCMeta):
    """
    Job service interface.
    """

    @abstractmethod
    async def create_job(self, id: bytes, name: str) -> JobModel:
        pass

    @abstractmethod
    async def run_component(
        self,
        request: JobComponentRequest,
        job_context: Optional[JobRunContext] = None,
    ) -> ComponentRunResult:
        """
        Run a single component for a job.

        :param request:
        :param job_context: Context of the job to run the component in.

        :return: A data structure describing the component run and resources
        modified/created/deleted.
        """

    @abstractmethod
    async def run_analyzer_by_attribute(
        self,
        request: JobAnalyzerRequest,
        job_context: Optional[JobRunContext] = None,
    ) -> ComponentRunResult:
        """
        Choose one or more Analyzer components to analyze the requested attributes on the given
        resource.

        :param request: Data structure containing the ID of the job to run the components in,
        the ID of the resource to run components on, which attributes the Analyzers should
        output, and the tags of the target resource.
        :param job_context: Context of the job to run the component in.

        :return: A data structure describing the component(s) run and resources
        modified/created/deleted.

        :raises NotFoundError: If no Analyzers can be found targeting the specified tags and
        outputting the specified attributes.
        """

    @abstractmethod
    async def run_components(
        self,
        request: JobMultiComponentRequest,
    ) -> ComponentRunResult:
        """
        Automatically select one or more components to run on a resource. The components must
        match the provided component filters and target at least one of the tags of the resource.

        :param request: Data structure containing the ID of the job to run the components in,
        the ID of the resource to run components on, and filters for the components to
        run.

        :return: A data structure describing the components run and resources
        modified/created/deleted.

        :raises ComponentAutoRunFailure: if one of the automatically chosen components raises an
        error while running.
        :raises NoMatchingComponentException: if no components match the filters for the resource.
        """

    @abstractmethod
    async def run_components_recursively(
        self,
        request: JobMultiComponentRequest,
    ) -> ComponentRunResult:
        """
        Start from a resource and run components on it and then on any resources which have tags
        added as a result of that initial run, then run components on any resources with new tags
        from those subsequent runs, until an iteration of component runs results in no new tags
        being added. The component(s) run on each resource are chosen according to the provided
        filters and which tags were added to that resource in the previous iteration. That is,
        the filters are applied to the set of resource which target those new tags.

        :param request: Data structure containing the ID of the job to run the components in,
        the ID of the resource to start running recursively from, and filters for the components to
        run.

        :return: A data structure describing the components run and resources
        modified/created/deleted.

        :raises ComponentAutoRunFailure: if one of the automatically chosen components raises an
        error while running.
        """

    @abstractmethod
    async def pack_recursively(
        self,
        job_id: bytes,
        resource_id: bytes,
    ) -> ComponentRunResult:
        """
        Call Packer components on the deepest descendants of a resource (the root of this search),
        then Packers on the next level up, etc. until the search root resource.

        :param job_id: Job to run the component in.
        :param resource_id: ID of the search root resource.

        :return: A data structure describing the components run and resources
        modified/created/deleted.
        """


class ComponentAutoRunFailure(Exception):
    def __init__(
        self,
        target_resource_id: bytes,
        component_filters: Tuple[ComponentFilter, ...],
        failing_component_id: bytes,
    ):
        super().__init__(
            f"Component {failing_component_id.decode()} failed when running on "
            f"{target_resource_id.hex()}. Component was chosen because it matched filters "
            f"{component_filters}"
        )
        self.target_resource_id = target_resource_id
        self.component_filters = component_filters
        self.failing_component_id = failing_component_id



import os
from dataclasses import dataclass
from typing import Tuple, List

import pytest

import test_ofrak.components

from ofrak import OFRAKContext
from ofrak.core.data import ReferencedDataAttributes


@dataclass
class DataRefsAnalyzerTestCase:
    """
    Test cases for ``DataRefsAnalyzerTestPattern``

    :ivar binary: path to binary to analyze
    :ivar expected_references: Expected analyzed references from func address to data address

    """

    binary: str
    expected_references: List[Tuple[int, int]]


DATA_REFS_TEST_CASES = [
    DataRefsAnalyzerTestCase(
        os.path.join(test_ofrak.components.ASSETS_DIR, "hello.out"),
        [(0x4004CC, 0x4005D8)],
    ),
    DataRefsAnalyzerTestCase(
        os.path.join(test_ofrak.components.ASSETS_DIR, "simple_arm_gcc.o.elf"),
        [(0x801C, 0x100B8)],
    ),
]


class DataRefsAnalyzerTestPattern:
    """
    Test pattern for testing DataRefsAnalyzer implementation for different frontends. Each test
    case should have at least all of the expected data refs analyzed successfully.
    """

    @pytest.fixture(params=DATA_REFS_TEST_CASES, ids=lambda tc: os.path.basename(tc.binary))
    async def data_refs_test_case(self, request) -> DataRefsAnalyzerTestCase:
        return request.param

    async def test_analyze_data_refs(
        self, ofrak_context: OFRAKContext, data_refs_test_case: DataRefsAnalyzerTestCase
    ):
        root_resource = await ofrak_context.create_root_resource_from_file(
            data_refs_test_case.binary
        )
        await root_resource.unpack()
        data_refs = await root_resource.analyze(ReferencedDataAttributes)
        await self.validate_data_refs(data_refs, data_refs_test_case)

    async def validate_data_refs(self, data_refs, test_case: DataRefsAnalyzerTestCase):
        assert 0 < len(data_refs.references)
        xrefs_to = data_refs.get_xrefs_to()
        xrefs_from = data_refs.get_xrefs_from()
        for expected_ref in test_case.expected_references:
            from_vaddr, to_data = expected_ref

            assert from_vaddr in xrefs_to[to_data]
            assert to_data in xrefs_from[from_vaddr]



from typing import Type, Any

import inspect
import sys
from typing_inspect import get_origin

from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


def is_metaclass(type_hint):
    """Will recognize Type, Type[X], and metaclasses"""
    return any(
        [
            type_hint == Type,
            get_origin(type_hint) == type,
            inspect.isclass(type_hint) and issubclass(type_hint, type),
        ]
    )


class TypeSerializer(SerializerInterface):
    """
    Serialize and deserialize classes (not instances) into `PJSONType`

    Implementation: the module path and name of the class are encoded into a single string.
    """

    targets = (is_metaclass,)

    def obj_to_pjson(self, cls: Type, _type_hint: Any) -> str:
        module = inspect.getmodule(cls)
        if module is None:
            raise TypeError(f"Can't find the module where {cls} was defined")
        import_path = module.__name__
        cls_name = cls.__name__
        return f"{import_path}.{cls_name}"

    def pjson_to_obj(self, pjson_obj: str, _type_hint: Any) -> Type:
        module_path, cls_name = pjson_obj.rsplit(".", maxsplit=1)
        # To avoid executing arbitrary code, only allow deserialization from modules
        # which are already loaded.
        try:
            module = sys.modules[module_path]
        except KeyError:
            raise ValueError(f"Can't deserialize {pjson_obj}: module not already loaded")
        cls = getattr(module, cls_name)
        return cls



import os.path
from dataclasses import dataclass
from typing import List, Dict

import pytest

from ofrak import OFRAKContext
from ofrak.core.basic_block import BasicBlock
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.filesystem import File
from ofrak.core.program import Program
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort
from pytest_ofrak.patterns import TEST_PATTERN_ASSETS_DIR
from pytest_ofrak.patterns.unpack_verify import (
    UnpackAndVerifyPattern,
    UnpackAndVerifyTestCase,
)
from ofrak_type.range import Range


@dataclass
class CodeRegionUnpackerTestCase(UnpackAndVerifyTestCase[int, List[int]]):
    binary_filename: str
    binary_md5_digest: str


CODE_REGION_UNPACKER_TEST_CASES = [
    CodeRegionUnpackerTestCase(
        "x64",
        {
            ## Required cases
            0x40040C: [],  # .text:call_gmon_start
            0x400430: [],  # .text:__do_global_dtors_aux
            0x4004A0: [],  # .text:frame_dummy
            0x4004C4: [],  # .text:main
            0x4004E0: [],  # .text:__libc_csu_fini
            0x4004F0: [],  # .text:__libc_csu_init
            0x400580: [],  # .text:__do_global_ctors_aux
        },
        {
            ## Optional cases
            0x400390: [],  # .init:_init (loader import)
            0x4003A8: [],  # .plt:LAB_004003a8 (trampoline, ignored if in .plt)
            0x4003B8: [],  # .plt:puts (lib import thunk)
            0x4003C8: [],  # .plt:__libc_start_main (lib import thunk)
            0x4003E0: [],  # .text:_start (loader import)
            0x4005B8: [],  # .fini:_fini (loader import)
        },
        "hello.out",
        "cc2de3c0cd2d0ded7543682c2470fcf0",
    ),
    CodeRegionUnpackerTestCase(
        "x64 ELF nosections",
        {
            ## Required cases
            0x40040C: [],  # .text:call_gmon_start
            0x400430: [],  # .text:__do_global_dtors_aux
            0x4004A0: [],  # .text:frame_dummy
            0x4004C4: [],  # .text:main
            0x4004F0: [],  # .text:__libc_csu_init
            0x400580: [],  # .text:__do_global_ctors_aux
        },
        {
            ## Optional cases
            0x4003E0: [],  # .text:_start (loader import)
            0x400390: [],  # .init:_init (loader import)
            0x4003A8: [],  # .plt:? (unknown import thunk)
            0x4003B8: [],  # .plt:puts (lib import thunk)
            0x4003C8: [],  # .plt:__libc_start_main (lib import thunk)
            0x4004E0: [],  # .text:__libc_csu_fini (handled ambiguously without .text hint)
            0x4005B8: [],  # .fini:_fini (loader import)
        },
        "hello_nosections.out",
        "56662f638390cae92c2cf5107bc3f1ef",
    ),
    CodeRegionUnpackerTestCase(
        "ARM ELF without literal pools",
        {
            0x8000: [],
            0x8030: [],
            0x8060: [],
            0x8094: [],
            0x8104: [],
        },
        {
            # No optional results
        },
        "arm_reloc_relocated.elf",
        "ed69056d3dbca810fa3a3f93db9e8927",
    ),
    CodeRegionUnpackerTestCase(
        "ARM ELF with literal pools",
        {
            0x8018: [0x8030],
            0x8034: [0x8060, 0x8064],
            0x8068: [],
        },
        {
            # .init
            0x8000: [],
            # .fini
            0x8090: [],
        },
        "simple_arm_gcc.o.elf",
        "c79d1bea0398d7a9d0faa1ba68786f5e",
    ),
]


class CodeRegionUnpackAndVerifyPattern(UnpackAndVerifyPattern):
    """
    Test pattern which checks a CodeRegionUnpacker implementation. This pattern is ready to go
    off-the-shelf. All that is needed to use it is
     1) A subclass with the name prefixed with "Test", and
     2) That subclass should be in a pytest context where the supplied frak context will have a
     CodeRegionUnpacker implementation.

    This file includes test cases of the type CodeRegionUnpackerTestCase. These include a URL
    which points to a binary which serves as the root resource. This binary will be unpacked
    recursively down to BasicBlocks, and all ComplexBlocks in the .text section will be
    extracted. Those ComplexBlocks should line up with the expected complex block virtual
    addresses, and if there are any data literals expected in the ComplexBlock those addresses are
    checked as well.
    """

    @pytest.fixture(params=CODE_REGION_UNPACKER_TEST_CASES, ids=lambda tc: tc.label)
    async def unpack_verify_test_case(self, request) -> CodeRegionUnpackerTestCase:
        return request.param

    @pytest.fixture
    async def root_resource(
        self,
        unpack_verify_test_case: CodeRegionUnpackerTestCase,
        ofrak_context: OFRAKContext,
        test_id: str,
    ) -> Resource:
        asset_path = os.path.join(TEST_PATTERN_ASSETS_DIR, unpack_verify_test_case.binary_filename)
        with open(asset_path, "rb") as f:
            binary_data = f.read()
        resource = await ofrak_context.create_root_resource(test_id, binary_data, tags=(File,))
        return resource

    async def unpack(self, root_resource: Resource):
        await root_resource.unpack_recursively(do_not_unpack=(BasicBlock,))

    async def get_descendants_to_verify(self, unpacked_resource: Resource) -> Dict[int, Resource]:
        program = await unpacked_resource.view_as(Program)
        code_regions = await program.get_code_regions()

        complex_blocks: List[ComplexBlock] = []

        for code_region in code_regions:
            complex_blocks += await code_region.resource.get_descendants_as_view(
                ComplexBlock,
                r_filter=ResourceFilter.with_tags(ComplexBlock),
                r_sort=ResourceSort(ComplexBlock.VirtualAddress),
            )

        return {cb.virtual_address: cb for cb in complex_blocks}

    async def verify_descendant(self, complex_block: ComplexBlock, specified_result: List[int]):
        start_address = complex_block.virtual_address
        end_address = start_address + complex_block.size
        # Check literal pools
        for lp_address in specified_result:
            assert lp_address in Range(
                start_address, end_address
            ), f"Literal pool at 0x{lp_address:x} not in extracted ComplexBlock {complex_block}"



import struct
from collections import defaultdict

from ofrak.service.id_service_i import IDServiceInterface


class SequentialIDService(IDServiceInterface):
    """
    An ID service implementation that generates sequencial ids to aid debugging.

    This ID service is intended to be used for debugging purposes only, since it relies on its
    internal state to generate ids.
    """

    def __init__(self):
        self._ids = defaultdict(lambda: 0)
        self._scope = "default"

    def generate_id(self) -> bytes:
        _id = self._ids[self._scope]
        self._ids[self._scope] += 1
        return struct.pack(">I", _id)

    @staticmethod
    def generate_id_from_base(base_id: bytes, key: str) -> bytes:
        return b"-".join((base_id, bytes(key, "utf-8")))



import logging
import math
from typing import Callable, List, Optional


def entropy_py(
    data: bytes, window_size: int, log_percent: Optional[Callable[[int], None]] = None
) -> bytes:
    """
    Return a list of entropy values where each value represents the Shannon entropy of the byte
    value distribution over a fixed-size, sliding window.
    """
    if log_percent is None:
        log_percent = lambda x: None
    else:
        # Sort of hacky way to know we are being called from the tests and don't need to log this
        logging.warning(
            f"Using the Python implementation of the Shannon entropy calculation! This is potentially "
            f"very slow, and is only used when the C extension cannot be built/found."
        )

    # Create a histogram, and populate it with initial values
    histogram = [0] * 256
    for b in data[:window_size]:
        histogram[b] += 1

    # Calculate the entropy using a sliding window
    entropy = [0] * (len(data) - window_size)
    last_percent_logged = 0
    for i in range(len(entropy)):
        entropy[i] = math.floor(255 * _shannon_entropy(histogram, window_size))
        histogram[data[i]] -= 1
        histogram[data[i + window_size]] += 1
        percent = int((i * 100) / len(data))
        if percent > last_percent_logged and percent % 10 == 0:
            log_percent(percent)
            last_percent_logged = percent
    return bytes(entropy)


def _shannon_entropy(distribution: List[int], window_size: int) -> float:
    """
    Return the Shannon entropy of the input probability distribution (represented as a histogram
    counting byte occurrences over a window of known size).

    Shannon entropy represents how uniform a probability distribution is. Since more uniform
    implies less predictable (because the probability of any outcome is equally likely in a
    uniform distribution), a sample with higher entropy is "more random" than one with lower
    entropy. More here: <https://en.wikipedia.org/wiki/Entropy_(information_theory)>.
    """

    result = 0.0
    for num_occurrences in distribution:
        probability = num_occurrences / window_size
        # Note that the zero check is required because the domain of log2 is the positive reals
        result += probability * math.log2(probability) if probability != 0.0 else 0.0
    return -result / math.log2(window_size)



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Set, Dict

from ofrak.model.tag_model import ResourceTag


@dataclass
class JobRunResourceTracker:
    tags_added: Set[ResourceTag] = field(default_factory=set)


@dataclass
class JobRunContext:
    trackers: Dict[bytes, JobRunResourceTracker] = field(
        default_factory=lambda: defaultdict(JobRunResourceTracker)
    )


class JobRunContextFactory:
    def create(self) -> JobRunContext:
        return JobRunContext()


@dataclass
class JobModel:
    id: bytes
    name: str



import logging
from dataclasses import dataclass
from typing import Dict, Callable, Union, Iterable

import magic

from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.core.binary import GenericBinary, GenericText
from ofrak.core.filesystem import File
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.tag_model import ResourceTag
from ofrak.resource import Resource
from ofrak_type.error import AlreadyExistError

LOGGER = logging.getLogger(__name__)


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Magic(ResourceAttributes):
    mime: str
    descriptor: str


class MagicAnalyzer(Analyzer[None, Magic]):
    """
    Analyze a binary blob to extract its mimetype and magic description.
    """

    targets = (File, GenericBinary)
    outputs = (Magic,)

    async def analyze(self, resource: Resource, config=None) -> Magic:
        data = await resource.get_data()
        magic_mime = magic.from_buffer(data, mime=True)
        magic_description = magic.from_buffer(data)
        return Magic(magic_mime, magic_description)


class MagicMimeIdentifier(Identifier[None]):
    """
    Identify and add the appropriate tag for a given resource based on its mimetype.
    """

    id = b"MagicMimeIdentifier"
    targets = (File, GenericBinary)
    _tags_by_mime: Dict[str, ResourceTag] = dict()

    async def identify(self, resource: Resource, config=None):
        _magic = await resource.analyze(Magic)
        magic_mime = _magic.mime
        tag = MagicMimeIdentifier._tags_by_mime.get(magic_mime)
        if tag is not None:
            resource.add_tag(tag)

    @classmethod
    def register(cls, resource: ResourceTag, mime_types: Union[Iterable[str], str]):
        if isinstance(mime_types, str):
            mime_types = [mime_types]
        for mime_type in mime_types:
            if mime_type in cls._tags_by_mime:
                raise AlreadyExistError(f"Registering already-registered mime type: {mime_type}")
            cls._tags_by_mime[mime_type] = resource


class MagicDescriptionIdentifier(Identifier[None]):
    """
    Identify and add the appropriate tag for a given resource based on its mime description.
    """

    id = b"MagicDescriptionIdentifier"
    targets = (File, GenericBinary)
    _matchers: Dict[Callable, ResourceTag] = dict()

    async def identify(self, resource: Resource, config):
        _magic = await resource.analyze(Magic)
        magic_description = _magic.descriptor
        for matcher, resource_type in self._matchers.items():
            if matcher(magic_description):
                resource.add_tag(resource_type)

    @classmethod
    def register(cls, resource: ResourceTag, matcher: Callable):
        if matcher in cls._matchers:
            raise AlreadyExistError("Registering already-registered matcher")
        cls._matchers[matcher] = resource


MagicMimeIdentifier.register(GenericText, "text/plain")
MagicDescriptionIdentifier.register(
    GenericText, lambda desc: any([("ASCII text" in s) for s in desc.split(", ")])
)

MagicMimeIdentifier.register(GenericBinary, "application/octet-stream")
MagicDescriptionIdentifier.register(GenericBinary, lambda s: s == "data")



from abc import ABC, abstractmethod
from typing import Tuple, Optional, List

from ofrak.component.abstract import AbstractComponent
from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.modifier import Modifier
from ofrak.model.component_filters import (
    ComponentWhitelistFilter,
    ComponentTypeFilter,
    ComponentOrMetaFilter,
    ComponentAndMetaFilter,
    ComponentNotMetaFilter,
)
from ofrak.model.component_model import CC
from ofrak.model.tag_model import ResourceTag
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.component_locator_i import (
    ComponentLocatorInterface,
)
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface


class UnpackerError(RuntimeError):
    """Base exception raised by unpackers."""


class Unpacker(AbstractComponent[CC], ABC):
    """
    Unpackers are components that unpack resources, splitting them into one or more children.
    """

    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        component_locator: ComponentLocatorInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service)
        self._component_locator = component_locator

    @property
    @abstractmethod
    def children(self) -> Tuple[Optional[ResourceTag], ...]:
        """
        A list of [ResourceTags][ofrak.model.tag_model.ResourceTag] that an unpacker can unpack a
        resource into.
        """
        raise NotImplementedError()

    @abstractmethod
    async def unpack(self, resource: Resource, config: CC) -> None:
        """
        Unpack the given resource.

        Users should not call this method directly; rather, they should run
        [Resource.run][ofrak.resource.Resource.run] or
        [Resource.unpack][ofrak.resource.Resource.unpack].

        :param resource: The resource that is being unpacked
        :param config: Optional config for unpacking. If an implementation provides a default,
        this default will always be used when config would otherwise be None. Note that a copy of
        the default config will be passed, so the default config values cannot be modified
        persistently by a component run.
        """
        raise NotImplementedError()

    @classmethod
    def get_default_config(cls) -> Optional[CC]:
        return cls._get_default_config_from_method(cls.unpack)

    async def _run(self, resource: Resource, config: CC):
        if resource.has_component_run(self.get_id(), self.get_version()):
            return self._log_component_has_run_warning(resource)
        if resource.has_component_run(self.get_id()):
            self._log_component_has_run_warning(resource)
            raise NotImplementedError(
                "If the component has already run (but on a different "
                "version), we should remove the dependencies"
            )
        await self.unpack(resource, config)
        resource.add_component(self.get_id(), self.get_version())
        self._validate_unpacked_children(resource)
        # Identify which packers ran (if any) and clear that record, so that it will be allowed
        # to run again
        packer_ids = self._get_which_packers_ran(resource)
        for packer_id in packer_ids:
            resource.remove_component(packer_id)

    def _get_which_packers_ran(self, resource: Resource) -> Tuple[bytes, ...]:
        unpackers_ran = self._component_locator.get_components_matching_filter(
            ComponentAndMetaFilter(
                ComponentWhitelistFilter(*resource.get_model().component_versions.keys()),
                # Use process of elimination to avoid circular import between unpacker.py, packer.py
                ComponentNotMetaFilter(
                    ComponentOrMetaFilter(
                        ComponentTypeFilter(Unpacker),  # type: ignore
                        ComponentTypeFilter(Analyzer),  # type: ignore
                        ComponentTypeFilter(Identifier),  # type: ignore
                        ComponentTypeFilter(Modifier),  # type: ignore
                    )
                ),
            )
        )
        return tuple(unpacker.get_id() for unpacker in unpackers_ran)

    def _validate_unpacked_children(self, resource: Resource) -> None:
        """
        Validate that the unpacked resources match the type defined by
        [Unpacker.children][ofrak.component.unpacker.Unpacker.children].

        :param resource:
        :raises ValueError: if the unpacked child does not match the type defined in
          [Unpacker.children][ofrak.component.unpacker.Unpacker.children]
        """
        component_context = resource.get_component_context()
        resource_context = resource.get_resource_context()
        untagged_descendants_allowed = None in self.children
        for descendant_id in component_context.resources_created:
            descendant_model = resource_context.resource_models.get(descendant_id)
            if descendant_model is None:
                raise ValueError(
                    f"Cannot find descendant {descendant_id.decode()} for resource "
                    f"{resource.get_id().decode()}."
                )
            descendant_has_tags = 0 != len(descendant_model.get_tags())

            if descendant_has_tags:
                if any(
                    descendant_model.has_tag(descendant_tag)
                    for descendant_tag in self.children
                    if descendant_tag is not None
                ):
                    continue
                else:
                    valid_tag_patterns: List[str] = [
                        descendant_tag.__name__
                        for descendant_tag in self.children
                        if descendant_tag is not None
                    ]
                    if untagged_descendants_allowed:
                        valid_tag_patterns.append("untagged resource")
                    expected_patterns = ", ".join(valid_tag_patterns)

                    raise ValueError(
                        f"Unpacker {self.get_id().decode()} created resource {descendant_id.hex()} "
                        f"but its tags {descendant_model.get_tags()} do not match any of the "
                        f"expected patterns this unpacker should create: {expected_patterns}"
                    )
            elif untagged_descendants_allowed:
                continue
            else:
                raise ValueError(
                    f"Unpacker {self.get_id().decode()} created resource {descendant_id.hex()} but "
                    f"its tags {descendant_model.get_tags()} do not match any of the expected "
                    f"patterns this unpacker should create: "
                    f"{', '.join([str(descendant_tag) for descendant_tag in self.children])}"
                )



from dataclasses import dataclass
from typing import Tuple, List

from ofrak.component.modifier import Modifier
from ofrak.core.memory_region import MemoryRegion
from ofrak.model.component_model import ComponentConfig
from ofrak.resource import Resource
from ofrak_type.range import Range


@dataclass
class BinaryInjectorModifierConfig(ComponentConfig):
    binary_patch_configs: List[Tuple[int, bytes]]


class BinaryInjectorModifier(Modifier[BinaryInjectorModifierConfig]):
    """
    Inject bytes at the given vm address in a [MemoryRegion][ofrak.core.memory_region.MemoryRegion].
    """

    targets = (MemoryRegion,)

    async def modify(self, resource: Resource, config: BinaryInjectorModifierConfig):
        assert config is not None
        memory_region = await resource.view_as(MemoryRegion)
        for vm_address, patch in config.binary_patch_configs:
            offset = memory_region.get_offset_in_self(vm_address)
            region_resource_data_id = memory_region.resource.get_data_id()
            if region_resource_data_id is None:
                raise ValueError(
                    "Cannot create DataPatch for a memory region resource with no " "data ID"
                )
            memory_region.resource.queue_patch(Range(offset, offset + len(patch)), patch)



from typing import Dict, Any

import json
import sys

import ofrak_type.error


class SerializedError(RuntimeError):
    def to_json(self):
        return json.dumps(self.to_dict(self))

    @classmethod
    def to_dict(cls, error: Exception):
        return {"type": type(error).__name__, "message": str(error)}

    @classmethod
    def from_json(cls, serialized: str) -> "SerializedError":
        error_dict = json.loads(serialized)
        error_type = error_dict["type"]
        try:
            error = getattr(sys.modules[__name__], error_type)
        except AttributeError:
            try:
                error = getattr(ofrak_type.error, error_type)
            except:
                raise ValueError(error_dict)
        if issubclass(error, cls):
            return error.from_dict({"message": error_dict["message"]})
        else:
            return error(error_dict["message"])

    @classmethod
    def from_dict(cls, error_dict: Dict[str, Any]) -> "SerializedError":
        return cls(error_dict["message"])


class DataServiceError(SerializedError):
    pass


class ResourceServiceError(SerializedError):
    pass


class JobServiceError(SerializedError):
    pass


class OutOfBoundError(DataServiceError):
    pass


class PatchOverlapError(DataServiceError):
    pass


class MisalignedError(DataServiceError):
    pass


class AmbiguousOrderError(RuntimeError):
    pass


class NonContiguousError(DataServiceError):
    pass



import asyncio
import functools
import logging
import json
import os
import sys
import webbrowser
from collections import defaultdict
from typing import (
    Iterable,
    Optional,
    Dict,
    cast,
    Set,
    Tuple,
    Union,
    Type,
    Callable,
    TypeVar,
    Any,
)

import orjson

from aiohttp import web
from aiohttp.web_exceptions import HTTPBadRequest
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from aiohttp.web_fileresponse import FileResponse

from ofrak.ofrak_context import get_current_ofrak_context
from ofrak_type.error import NotFoundError
from ofrak_type.range import Range

from ofrak import (
    OFRAKContext,
    ResourceFilter,
    ResourceAttributeRangeFilter,
    ResourceAttributeValueFilter,
    ResourceSort,
    ResourceTag,
)
from ofrak.core import Addressable, File
from ofrak.core import (
    GenericBinary,
    AddCommentModifier,
    AddCommentModifierConfig,
    DeleteCommentModifierConfig,
    DeleteCommentModifier,
    StringFindReplaceConfig,
    StringFindReplaceModifier,
)
from ofrak.model.component_model import (
    ComponentContext,
    ClientComponentContext,
    ComponentRunResult,
)
from ofrak.model.resource_model import (
    ResourceContext,
    ClientResourceContext,
    ResourceModel,
    ResourceAttributes,
    MutableResourceModel,
)
from ofrak.model.viewable_tag_model import ResourceViewContext
from ofrak.resource import Resource
from ofrak.service.error import SerializedError
from ofrak.service.serialization.pjson import (
    SerializationServiceInterface,
    PJSONSerializationService,
)
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.core.entropy import DataSummaryAnalyzer

T = TypeVar("T")
LOGGER = logging.getLogger(__name__)


def exceptions_to_http(error_class: Type[SerializedError]):
    """
    Decorator for a server function that attempts to do some work, and
    forwards the exception, if any, to the client over HTTP.

    Usage:

    @exceptions_to_http(MyErrorClass)
    async def handle_some_request(self, request...):
        ...
    """

    def exceptions_to_http_decorator(func: Callable):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as error:
                LOGGER.exception("Exception raised in aiohttp endpoint")
                return respond_with_error(error, error_class)

        return wrapper

    return exceptions_to_http_decorator


class AiohttpOFRAKServer:
    routes = web.RouteTableDef()

    def __init__(
        self,
        serializer: SerializationServiceInterface,
        ofrak_context: OFRAKContext,
        host: str,
        port: int,
    ):
        self._serializer = serializer
        self._app = web.Application(client_max_size=None)  # type: ignore
        self._host = host
        self._port = port
        self._ofrak_context = ofrak_context
        self.resource_context: ResourceContext = ClientResourceContext()
        self.resource_view_context: ResourceViewContext = ResourceViewContext()
        self.component_context: ComponentContext = ClientComponentContext()

        self._app.add_routes(
            [
                web.post("/create_root_resource", self.create_root_resource),
                web.get("/get_root_resources", self.get_root_resources),
                web.get("/{resource_id}/", self.get_resource),
                web.get("/{resource_id}/get_data", self.get_data),
                web.post(
                    "/batch/get_data_range_within_parent",
                    self.batch_get_range,
                ),
                web.get(
                    "/{resource_id}/get_child_data_ranges",
                    self.get_child_data_ranges,
                ),
                web.get("/{resource_id}/get_root", self.get_root_resource_from_child),
                web.post("/{resource_id}/unpack", self.unpack),
                web.post("/{resource_id}/unpack_recursively", self.unpack_recursively),
                web.post("/{resource_id}/pack", self.pack),
                web.post("/{resource_id}/pack_recursively", self.pack_recursively),
                web.post("/{resource_id}/analyze", self.analyze),
                web.post("/{resource_id}/identify", self.identify),
                web.post("/{resource_id}/data_summary", self.data_summary),
                web.get("/{resource_id}/get_parent", self.get_parent),
                web.get("/{resource_id}/get_ancestors", self.get_ancestors),
                web.post("/batch/get_children", self.batch_get_children),
                web.post("/{resource_id}/queue_patch", self.queue_patch),
                web.post("/{resource_id}/create_mapped_child", self.create_mapped_child),
                web.post("/{resource_id}/find_and_replace", self.find_and_replace),
                web.post("/{resource_id}/add_comment", self.add_comment),
                web.post("/{resource_id}/delete_comment", self.delete_comment),
                web.post("/{resource_id}/search_for_vaddr", self.search_for_vaddr),
                web.post("/{resource_id}/add_tag", self.add_tag),
                web.get("/get_all_tags", self.get_all_tags),
                web.get("/", self.get_static_files),
                web.static(
                    "/",
                    os.path.join(os.path.dirname(__file__), "./public"),
                    show_index=True,
                ),
            ]
        )

        self._job_ids: Dict[str, bytes] = defaultdict(
            lambda: ofrak_context.id_service.generate_id()
        )

    async def start(self):  # pragma: no cover
        """
        Start the server then return.
        """
        self.runner = web.AppRunner(self._app)
        await self.runner.setup()
        server = web.TCPSite(self.runner, host=self._host, port=self._port)
        await server.start()

    async def stop(self):  # pragma: no cover
        """
        Stop the server.
        """
        await self.runner.server.shutdown()
        await self.runner.cleanup()

    async def run_until_cancelled(self):  # pragma: no cover
        """
        To be run after `start_server`, within an asyncio Task.
        cancel() that task to shutdown the server.
        """
        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            pass
        finally:
            await self.runner.cleanup()

    @exceptions_to_http(SerializedError)
    async def create_root_resource(self, request: Request) -> Response:
        name = request.query.get("name")
        if name is None:
            return HTTPBadRequest(reason="Missing root resource `name` from request")
        resource_data = await request.read()
        root_resource = await self._ofrak_context.create_root_resource(name, resource_data, (File,))
        if request.remote is not None:
            self._job_ids[request.remote] = root_resource.get_job_id()
        return json_response(self._serialize_resource(root_resource))

    @exceptions_to_http(SerializedError)
    async def get_root_resources(self, request: Request) -> Response:
        roots = await self._ofrak_context.resource_service.get_root_resources()
        return json_response(list(map(self._serialize_resource_model, roots)))

    @exceptions_to_http(SerializedError)
    async def get_resource(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def get_data(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        _range = self._serializer.from_pjson(
            get_query_string_as_pjson(request).get("range"), Optional[Range]
        )
        data = await resource.get_data(_range)
        return Response(body=data)

    @exceptions_to_http(SerializedError)
    async def get_child_data_ranges(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        resource_service = self._ofrak_context.resource_factory._resource_service
        data_service = self._ofrak_context.resource_factory._data_service
        children = await resource_service.get_descendants_by_id(
            resource.get_id(),
            max_depth=1,
        )

        async def get_range(child):
            try:
                if child.data_id is None:
                    return
                data_range = await data_service.get_range_within_other(
                    child.data_id, resource.get_data_id()
                )
                return child.id.hex(), (data_range.start, data_range.end)
            except ValueError:
                pass

        return json_response(
            dict(filter(lambda x: x is not None, await asyncio.gather(*map(get_range, children))))
        )

    @exceptions_to_http(SerializedError)
    async def batch_get_range(self, request: Request) -> Response:
        if request.remote is not None:
            job_id = self._job_ids[request.remote]
        else:
            raise ValueError("No IP address found for the remote request!")

        async def get_resource_range(resource_id):
            resource_model = await self._get_resource_model_by_id(
                bytes.fromhex(resource_id), job_id
            )
            if resource_model.data_id is None:
                raise ValueError(
                    "Resource does not have a data_id. Cannot get data range from a "
                    "resource with no data."
                )
            if resource_model.parent_id is None:
                data_range = Range(0, 0)
            else:
                resource_service = self._ofrak_context.resource_factory._resource_service
                data_service = self._ofrak_context.resource_factory._data_service
                parent_models = list(
                    await resource_service.get_ancestors_by_id(resource_model.id, max_count=1)
                )
                if len(parent_models) != 1:
                    raise NotFoundError(
                        f"There is no parent for resource {resource_model.id.hex()}"
                    )
                parent_model = parent_models[0]

                parent_data_id = parent_model.data_id
                if parent_data_id is None:
                    data_range = Range(0, 0)
                else:
                    try:
                        data_range = await data_service.get_range_within_other(
                            resource_model.data_id, parent_data_id
                        )
                    except ValueError:
                        data_range = Range(0, 0)
            return resource_id, [data_range.start, data_range.end]

        return json_response(
            dict(await asyncio.gather(*map(get_resource_range, await request.json())))
        )

    @exceptions_to_http(SerializedError)
    async def unpack(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        result = await resource.unpack()
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def unpack_recursively(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        result = await resource.unpack_recursively()
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def pack(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        result = await resource.pack()
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def pack_recursively(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        result = await resource.pack_recursively()
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def identify(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        result = await resource.identify()
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def data_summary(self, request: Request) -> Response:
        resource = cast(Resource, await self._get_resource_for_request(request))
        result = await resource.run(DataSummaryAnalyzer)
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def analyze(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        result = await resource.auto_run(all_analyzers=True)
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    @exceptions_to_http(SerializedError)
    async def get_parent(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        parent = await resource.get_parent()
        return json_response(self._serialize_resource(parent))

    @exceptions_to_http(SerializedError)
    async def get_ancestors(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        # TODO: filter argument
        ancestors = await resource.get_ancestors()
        return json_response(self._serialize_multi_resource(ancestors))

    @exceptions_to_http(SerializedError)
    async def batch_get_children(self, request: Request) -> Response:
        if request.remote is not None:
            job_id = self._job_ids[request.remote]
        else:
            raise ValueError("No IP address found for the remote request!")

        async def get_resource_children(resource_id):
            resource = await self._get_resource_by_id(bytes.fromhex(resource_id), job_id)
            child_models = await resource._resource_service.get_descendants_by_id(
                resource._resource.id,
                max_depth=1,
            )
            serialized_children = list(map(self._serialize_resource_model, child_models))
            serialized_children.sort(key=get_child_sort_key)

            return resource_id, serialized_children

        def get_child_sort_key(child):
            attrs = dict(child.get("attributes"))
            data_attr = attrs.get("ofrak.model.resource_model.Data")
            if data_attr is not None:
                return data_attr[1]["_offset"]
            else:
                return sys.maxsize

        return json_response(
            dict(await asyncio.gather(*map(get_resource_children, await request.json())))
        )

    @exceptions_to_http(SerializedError)
    async def get_root_resource_from_child(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        parent = resource
        try:
            # Assume get_ancestors returns an ordered list with the parent first and the root last
            for parent in await resource.get_ancestors():
                pass
        except NotFoundError:
            pass
        return json_response(self._serialize_resource(parent))

    @exceptions_to_http(SerializedError)
    async def queue_patch(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        new_data = await request.read()

        start_param = request.query.get("start")
        start = int(start_param) if start_param is not None else 0
        end_param = request.query.get("end")
        end = int(end_param) if end_param is not None else (await resource.get_data_length())

        resource.queue_patch(Range(start, end), new_data)
        await resource.save()
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def create_mapped_child(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        _range = self._serializer.from_pjson(await request.json(), Optional[Range])
        child = await resource.create_child(tags=(GenericBinary,), data_range=_range)
        return json_response(self._serialize_resource(child))

    @exceptions_to_http(SerializedError)
    async def find_and_replace(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        config = self._serializer.from_pjson(await request.json(), StringFindReplaceConfig)
        result = await resource.run(StringFindReplaceModifier, config=config)
        response_pjson = await self._serialize_component_result(result)
        return json_response(response_pjson)

    async def add_comment(self, request: Request) -> Response:
        """
        Expected POST body is a comment in the form Tuple[Optional[Range], str] (serialized to JSON).
        """
        resource = await self._get_resource_for_request(request)
        comment = self._serializer.from_pjson(await request.json(), Tuple[Optional[Range], str])
        result = await resource.run(AddCommentModifier, AddCommentModifierConfig(comment))
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def delete_comment(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        comment_range = self._serializer.from_pjson(await request.json(), Optional[Range])
        result = await resource.run(
            DeleteCommentModifier, DeleteCommentModifierConfig(comment_range)
        )
        return json_response(await self._serialize_component_result(result))

    @exceptions_to_http(SerializedError)
    async def search_for_vaddr(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        vaddr_start, vaddr_end = self._serializer.from_pjson(
            await request.json(), Tuple[int, Optional[int]]
        )
        try:
            vaddr_filter: Union[ResourceAttributeRangeFilter, ResourceAttributeValueFilter]
            if vaddr_end is not None:
                vaddr_filter = ResourceAttributeRangeFilter(
                    Addressable.VirtualAddress, vaddr_start, vaddr_end
                )
            else:
                vaddr_filter = ResourceAttributeValueFilter(Addressable.VirtualAddress, vaddr_start)
            matching_resources = await resource.get_descendants(
                r_filter=ResourceFilter(attribute_filters=(vaddr_filter,)),
                r_sort=ResourceSort(Addressable.VirtualAddress),
            )
            return json_response(list(map(self._serialize_resource, matching_resources)))

        except NotFoundError:
            return json_response([])

    @exceptions_to_http(SerializedError)
    async def add_tag(self, request: Request) -> Response:
        resource = await self._get_resource_for_request(request)
        tag = self._serializer.from_pjson(await request.json(), ResourceTag)
        resource.add_tag(tag)
        await resource.save()
        return json_response(self._serialize_resource(resource))

    @exceptions_to_http(SerializedError)
    async def get_all_tags(self, request: Request) -> Response:
        return json_response(
            self._serializer.to_pjson(self._ofrak_context.get_all_tags(), Set[ResourceTag])
        )

    @exceptions_to_http(SerializedError)
    async def get_static_files(self, request: Request) -> FileResponse:
        return FileResponse(os.path.join(os.path.dirname(__file__), "./public/index.html"))

    async def _get_resource_by_id(self, resource_id: bytes, job_id: bytes) -> Resource:
        resource = await self._ofrak_context.resource_factory.create(
            job_id,
            resource_id,
            self.resource_context,
            self.resource_view_context,
            self.component_context,
        )
        return resource

    async def _get_resource_model_by_id(
        self, resource_id: bytes, job_id: bytes
    ) -> Optional[Union[ResourceModel, MutableResourceModel]]:
        resource_m: Optional[Union[ResourceModel, MutableResourceModel]] = None
        resource_m = self.resource_context.resource_models.get(resource_id)
        if resource_m is None:
            resource_m = await self._ofrak_context.resource_factory._resource_service.get_by_id(
                resource_id
            )
        return resource_m

    async def _serialize_component_result(self, result: ComponentRunResult) -> PJSONType:
        async def get_and_serialize(resource_id) -> PJSONType:
            resource_model = await self._ofrak_context.resource_service.get_by_id(resource_id)
            return self._serialize_resource_model(resource_model)

        serialized_result = {
            "created": await asyncio.gather(*map(get_and_serialize, result.resources_created)),
            "modified": await asyncio.gather(
                *map(
                    get_and_serialize,
                    result.resources_modified.difference(result.resources_created),
                )
            ),
            "deleted": self._serializer.to_pjson(result.resources_deleted, Set[bytes]),
        }
        return serialized_result

    async def _get_resource_for_request(self, request: Request) -> Resource:
        resource_id = pluck_id(request, "resource_id")
        if request.remote is not None:
            job_id = self._job_ids[request.remote]
        else:
            raise ValueError("No IP address found for the remote request!")
        return await self._get_resource_by_id(resource_id, job_id)

    def _serialize_resource_model(self, resource_model: ResourceModel) -> PJSONType:
        """
        Serialize the resource model, stripped of information irrelevant to the frontend.
        """
        result = {
            "id": resource_model.id.hex(),
            "data_id": resource_model.data_id.hex() if resource_model.data_id else None,
            "parent_id": resource_model.parent_id.hex() if resource_model.parent_id else None,
            "tags": [tag.__module__ + "." + tag.__qualname__ for tag in resource_model.tags],
            "attributes": self._serializer.to_pjson(
                resource_model.attributes, Dict[Type[ResourceAttributes], ResourceAttributes]
            ),
            "caption": resource_model.caption,
        }
        return result

    def _serialize_resource(self, resource: Resource) -> PJSONType:
        """
        Serialize the resource as a serialized model, stripped of information irrelevant to the
        frontend.
        """
        return self._serialize_resource_model(resource.get_model())

    def _serialize_multi_resource(self, resources: Iterable[Resource]) -> PJSONType:
        """
        Serialize the resources as serialized models, stripped of information irrelevant to the
        frontend.
        """
        return list(map(self._serialize_resource, resources))


async def start_server(
    ofrak_context: OFRAKContext, host: str, port: int
) -> AiohttpOFRAKServer:  # pragma: no cover
    # Force using the correct PJSON serialization with the expected structure. Otherwise the
    # dependency injector may accidentally use the Stashed PJSON serialization service,
    # which returns PJSON that has a different, problematic structure.
    ofrak_context.injector.bind_factory(PJSONSerializationService)

    ofrak_context.injector.bind_factory(
        AiohttpOFRAKServer,
        ofrak_context=ofrak_context,
        host=host,
        port=port,
    )
    server = await ofrak_context.injector.get_instance(AiohttpOFRAKServer)
    await server.start()

    return server


def respond_with_error(error: Exception, error_cls: Type[SerializedError]) -> Response:
    if isinstance(error, error_cls):
        text = error.to_json()
    else:
        text = json.dumps(error_cls.to_dict(error))
    response = Response(text=text, status=500)
    return response


def pluck_id(request: Request, get_parameter_name: str) -> bytes:
    return bytes.fromhex(request.match_info[get_parameter_name])


def get_query_string_as_pjson(request: Request) -> Dict[str, PJSONType]:
    """
    URL-encoded GET parameters are all strings. For example, None is encoded as 'None',
    or 1 as '1', which isn't valid PJSON. We fix this by applying `json.loads` on each parameter.
    """
    return {key: json.loads(value) for key, value in request.query.items()}


async def open_gui(
    host: str,
    port: int,
    focus_resource: Optional[Resource] = None,
    ofrak_context: Optional[OFRAKContext] = None,
    open_in_browser: bool = True,
) -> AiohttpOFRAKServer:  # pragma: no cover
    if ofrak_context is None:
        ofrak_context = get_current_ofrak_context()

    server = await start_server(ofrak_context, host, port)

    if focus_resource is None:
        url = f"http://{server._host}:{server._port}/"
    else:
        url = f"http://{server._host}:{server._port}/#{focus_resource.get_id().hex()}"
    print(f"GUI is being served on {url}")

    if open_in_browser:
        webbrowser.open(url)

    return server


def json_response(
    data: Any = None,
    *,
    text: Optional[str] = None,
    body: Optional[bytes] = None,
    status: int = 200,
    reason: Optional[str] = None,
    headers=None,
    content_type: str = "application/json",
    dumps=orjson.dumps,
) -> Response:
    if data is not None:
        if text or body:
            raise ValueError("only one of data, text, or body should be specified")
        else:
            body = dumps(data)
    return Response(
        text=text,
        body=body,
        status=status,
        reason=reason,
        headers=headers,
        content_type=content_type,
    )



import os
import stat
import tempfile
from abc import ABC, abstractmethod
from subprocess import CalledProcessError

try:
    import xattr
except ImportError:
    import ofrak.core.xattr_stub as xattr  # type: ignore[no-redef]

from ofrak import OFRAKContext
from ofrak.component.abstract import ComponentSubprocessError
from ofrak.resource import Resource


class FilesystemPackUnpackVerifyPattern(ABC):
    """
    Generic test pattern for filesystem packer/unpackers to ensure that no information is lost
    after packing and unpacking a filesystem of more than one file.
    """

    def setup(self):
        """
        Override to set values before the test is run. For example use this to change the value of
        self.check_xattrs in a subclass.
        """
        self.check_xattrs = True
        self.check_stat = True

    async def test_pack_unpack_verify(self, ofrak_context: OFRAKContext):
        try:
            self.setup()
            with tempfile.TemporaryDirectory() as root_path:
                self.create_local_file_structure(root_path)
                root_resource = await self.create_root_resource(ofrak_context, root_path)
                await self.unpack(root_resource)
                await self.repack(root_resource)
                with tempfile.TemporaryDirectory() as extract_dir:
                    await self.extract(root_resource, extract_dir)
                    self.verify_filesystem_equality(root_path, extract_dir)
        except CalledProcessError as e:
            # Better printing of errors if something goes wrong in test setup/execution
            raise ComponentSubprocessError(e)

    def _dirs_from_list(self, parent, names, depth):
        # Create a file
        if depth == 0:
            for filename in names:
                with open(os.path.join(parent, f"{filename}.txt"), "w") as f:
                    f.write(filename)
            return

        # Create another layer of directories
        for d in names:
            path = os.path.join(parent, d)
            os.mkdir(path)
            self._dirs_from_list(path, names, depth - 1)

    def create_local_file_structure(self, root: str):
        """
        Create a local file structure with multiple folders and files, many with different
        permissions and xattrs. Return the path to the root directory. Override this method to
        test with a different directory structure, or to add xattrs or stat values.
        """
        self._dirs_from_list(root, ["a", "b", "c", "d"], 3)

    @abstractmethod
    async def create_root_resource(self, ofrak_context: OFRAKContext, directory: str) -> Resource:
        """
        Create a packed version of the filesystem outside of OFRAK that it can test unpacking and
        repacking.
        """
        raise NotImplementedError()

    @abstractmethod
    async def unpack(self, root_resource: Resource):
        """
        Unpack the filesystem using the OFRAK unpackers that are being tested.
        """
        raise NotImplementedError()

    @abstractmethod
    async def repack(self, root_resource: Resource):
        """
        Repack the filesystem using the OFRAK packers that are being tested.
        """
        raise NotImplementedError()

    @abstractmethod
    async def extract(self, root_resource: Resource, extract_dir: str):
        """
        Flush the packed filesystem to disk and then extract it using a method outside of OFRAK
        so the resulting path can be compared with the original.
        """
        raise NotImplementedError()

    def verify_filesystem_equality(self, old_path: str, new_path: str):
        if self.check_stat:
            self._validate_stat_equality(old_path, new_path)
        if self.check_xattrs:
            self._validate_xattrs_equality(old_path, new_path)
        self._validate_type_equality(old_path, new_path)
        if os.path.islink(old_path):
            self._validate_link_equality(old_path, new_path)
        elif os.path.isfile(old_path):
            self._validate_file_data_equality(old_path, new_path)
        elif os.path.isdir(old_path):
            self._validate_folder_equality(old_path, new_path)
        else:
            self._validate_special_file_equality(old_path, new_path)

    def _validate_stat_equality(self, old_path: str, new_path: str):
        old_stat = os.lstat(old_path)
        new_stat = os.lstat(new_path)
        assert (
            old_stat == new_stat
        ), f"{old_path} and {new_path} have different stat values\nold: {old_stat}\nnew: {new_stat}"

    def _validate_xattrs_equality(self, old_path: str, new_path: str):
        old_xattrs = dict()
        for attr in xattr.listxattr(old_path, symlink=True):
            old_xattrs[attr] = xattr.getxattr(old_path, attr, symlink=True)
        new_xattrs = dict()
        for attr in xattr.listxattr(new_path, symlink=True):
            new_xattrs[attr] = xattr.getxattr(new_path, attr, symlink=True)
        assert (
            old_xattrs == new_xattrs
        ), f"{old_path} and {new_path} have different xattrs\nold: {old_xattrs}\nnew: {new_xattrs}"

    def _validate_type_equality(self, old_path: str, new_path: str):
        assert (
            os.path.isdir(old_path) == os.path.isdir(new_path)
            and os.path.isfile(old_path) == os.path.isfile(new_path)
            and os.path.islink(old_path) == os.path.islink(new_path)
            and os.path.ismount(old_path) == os.path.ismount(new_path)
        ), f"{old_path} and {new_path} are not the same type (file/folder/symlink)"

    def _validate_file_data_equality(self, old_path: str, new_path: str):
        with open(old_path, "rb") as old_f, open(new_path, "rb") as new_f:
            assert (
                old_f.read() == new_f.read()
            ), f"{old_path} and {new_path} do not contain the same data"

    def _validate_folder_equality(self, old_path: str, new_path: str):
        old_files = os.listdir(old_path)
        new_files = os.listdir(new_path)
        assert len(old_files) == len(new_files) and set(old_files) == set(
            new_files
        ), f"{old_path} and {new_path} contain different files\nold: {old_files}\nnew: {new_files}"

        for old_f, new_f in zip(old_files, new_files):
            old_f = os.path.join(old_path, old_f)
            new_f = os.path.join(new_path, new_f)
            self.verify_filesystem_equality(old_f, new_f)

    def _validate_link_equality(self, old_path: str, new_path: str):
        old_source = os.readlink(old_path)
        new_source = os.readlink(new_path)
        assert old_source == new_source, (
            f"{old_path} and {new_path} point to different files\n"
            f"old: {old_source}\nnew: {new_source}"
        )

    def _validate_special_file_equality(self, old_path: str, new_path: str):
        old_stat = os.lstat(old_path)
        new_stat = os.lstat(new_path)
        assert old_stat.st_mode == new_stat.st_mode, (
            f"{old_path} and {new_path} have different "
            f"modes ({stat.S_IFMT(old_stat.st_mode):o} vs {stat.S_IFMT(new_stat.st_mode):o})"
        )
        assert old_stat.st_uid == new_stat.st_uid, (
            f"{old_path} and {new_path} have different "
            f"uids ({old_stat.st_uid} vs {new_stat.st_uid})"
        )
        assert old_stat.st_gid == new_stat.st_gid, (
            f"{old_path} and {new_path} have different "
            f"gids ({old_stat.st_gid} vs {new_stat.st_gid})"
        )



from collections.abc import Sequence, Iterable
from typing import Any, List, Union

from beartype import beartype
from typing_inspect import get_origin, get_args

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class SequenceSerializer(SerializerInterface):
    """
    Serialize and deserialize `List[X]`, `Sequence[X]` and `Iterable[X]` into `PJSONType`.

    Implementation: all of these are serialized as lists.
    """

    targets = (lambda type_hint: get_origin(type_hint) in (list, Sequence, Iterable),)

    SupportedSequenceType = Union[List[Any], Iterable, Sequence]

    @beartype
    def obj_to_pjson(self, obj: SupportedSequenceType, type_hint: Any) -> List[PJSONType]:
        args = get_args(type_hint)
        return [self._service.to_pjson(item, args[0]) for item in obj]

    @beartype
    def pjson_to_obj(self, pjson_obj: List[PJSONType], type_hint: Any) -> SupportedSequenceType:
        args = get_args(type_hint)
        return [self._service.from_pjson(item, args[0]) for item in pjson_obj]



import dataclasses
import inspect
import logging
from abc import ABC, abstractmethod
from subprocess import CalledProcessError
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Callable,
    Any,
    cast,
    Tuple,
)

from ofrak.component.interface import ComponentInterface
from ofrak.model.component_model import (
    ComponentContext,
    CC,
    ComponentRunResult,
    ComponentConfig,
    ComponentExternalTool,
)
from ofrak.model.data_model import DataPatchesResult
from ofrak.model.job_model import (
    JobRunContext,
)
from ofrak.model.resource_model import (
    ResourceContext,
    MutableResourceModel,
)
from ofrak.model.viewable_tag_model import ResourceViewContext
from ofrak.resource import Resource, ResourceFactory, save_resources
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.dependency_handler import DependencyHandlerFactory
from ofrak.service.resource_service_i import ResourceServiceInterface
from ofrak_type.error import NotFoundError

LOGGER = logging.getLogger(__name__)


class AbstractComponent(ComponentInterface[CC], ABC):
    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
    ):
        self._resource_factory = resource_factory
        self._data_service = data_service
        self._resource_service = resource_service
        self._dependency_handler_factory = DependencyHandlerFactory()
        self._default_config = self.get_default_config()

    @classmethod
    def get_id(cls) -> bytes:
        return cls.id if cls.id is not None else cls.__name__.encode()

    # By default, assume component has no external dependencies
    external_dependencies: Tuple[ComponentExternalTool, ...] = ()

    async def run(
        self,
        job_id: bytes,
        resource_id: bytes,
        job_context: JobRunContext,
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        config: CC,
    ) -> ComponentRunResult:
        """

        :param job_id:
        :param resource_id:
        :param job_context:
        :param resource_context:
        :param resource_view_context:
        :param config:
        :return: The IDs of all resources modified by this component
        """
        component_context = ComponentContext(self.get_id(), self.get_version())
        resource = await self._resource_factory.create(
            job_id,
            resource_id,
            resource_context,
            resource_view_context,
            component_context,
            job_context,
        )
        if config is None and self._default_config is not None:
            config = dataclasses.replace(self._default_config)
        try:
            await self._run(resource, config)
        except FileNotFoundError as e:
            # Check if the problem was that one of the dependencies is missing
            missing_file = e.filename
            for dep in self.external_dependencies:
                if dep.tool == missing_file:
                    raise ComponentMissingDependencyError(self, dep)
            raise
        except CalledProcessError as e:
            raise ComponentSubprocessError(e)

        deleted_resource_models: List[MutableResourceModel] = list()

        for deleted_r_id in component_context.resources_deleted:
            mutable_resource_model = resource_context.resource_models.get(deleted_r_id)
            if mutable_resource_model:
                deleted_resource_models.append(mutable_resource_model)
            else:
                raise NotFoundError(
                    f"The resource {deleted_r_id.hex()} was deleted but not in "
                    f"the resource context"
                )

        # Save deleted resource so they won't interfere with patches
        # This is where deleted resources are actually deleted from their respective databases
        await self._save_resources(
            job_id,
            deleted_resource_models,
            resource_context,
            resource_view_context,
            job_context,
            component_context,
        )

        dependency_handler = self._dependency_handler_factory.create(
            self._resource_service,
            self._data_service,
            component_context,
            resource_context,
        )
        patch_results = await self.apply_all_patches(component_context)
        await dependency_handler.handle_post_patch_dependencies(patch_results)
        dependency_handler.create_component_dependencies(self.get_id(), self.get_version())
        dependency_handler.create_resource_dependencies(self.get_id())

        # Get modified resources
        modified_resource_models: Dict[bytes, MutableResourceModel] = dict()
        modified_resource_ids = component_context.get_modified_resource_ids()
        for modified_r_id in modified_resource_ids:
            mutable_resource_model = resource_context.resource_models.get(modified_r_id)
            if mutable_resource_model:
                modified_resource_models[modified_r_id] = mutable_resource_model
            else:
                raise NotFoundError(
                    f"The resource {modified_r_id.hex()} was modified but not in "
                    f"the resource context"
                )

        # Include resources modified by data patches in `modified_resource_ids`
        data_ids_to_models = await dependency_handler.map_data_ids_to_resources(
            patch_result.data_id for patch_result in patch_results
        )
        for m in data_ids_to_models.values():
            modified_resource_ids.add(m.id)

        # Exclude deleted resources from `modified_resource_ids`
        # (deleting and modifying are handled separately)
        modified_resource_ids.difference_update(component_context.resources_deleted)

        # Save modified resources
        await self._save_resources(
            job_id,
            modified_resource_models.values(),
            resource_context,
            resource_view_context,
            job_context,
            component_context,
        )

        component_result = ComponentRunResult(
            {self.get_id()},
            modified_resource_ids,
            component_context.resources_deleted,
            component_context.resources_created,
        )
        return component_result

    @abstractmethod
    async def _run(self, resource: Resource, config: CC):
        raise NotImplementedError()

    async def _save_resources(
        self,
        job_id: bytes,
        mutable_resource_models: Iterable[MutableResourceModel],
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        job_context: Optional[JobRunContext],
        component_context: ComponentContext,
    ):
        resources = await self._resource_factory.create_many(
            job_id,
            (mutable_resource_model.id for mutable_resource_model in mutable_resource_models),
            resource_context,
            resource_view_context,
            component_context,
            job_context,
        )
        await save_resources(
            resources,
            self._resource_service,
            self._data_service,
            component_context,
            resource_context,
            resource_view_context,
        )

    @staticmethod
    def _get_default_config_from_method(
        component_method: Callable[[Any, Resource, CC], Any]
    ) -> Optional[CC]:
        run_signature = inspect.signature(component_method)
        config_arg_type = run_signature.parameters["config"]
        default_arg: CC = config_arg_type.default

        if isinstance(default_arg, ComponentConfig):
            try:
                return cast(CC, default_arg)
            except TypeError as e:
                raise TypeError(
                    f"ComponentConfig subclass {type(default_arg)} is not a dataclass! This is "
                    f"required in order to copy the default config to ensure the default is "
                    f"non-mutable."
                ) from e
        elif default_arg is not None and default_arg is not config_arg_type.empty:
            raise TypeError(
                f"Default config {default_arg} must be either an instance of ComponentConfig, "
                f"None, or left empty!"
            )
        else:
            return None

    async def apply_all_patches(
        self, component_context: ComponentContext
    ) -> List[DataPatchesResult]:
        # Build a list of patches, making sure that there is at most one patch that causes a resize
        patches = []
        for resource_id, tracker in component_context.modification_trackers.items():
            patches.extend(tracker.data_patches)
            tracker.data_patches.clear()
        if len(patches) > 0:
            return await self._data_service.apply_patches(patches)
        else:
            return []

    def get_version(self) -> int:
        return 1

    def _log_component_has_run_warning(self, resource: Resource):
        LOGGER.warning(
            f"{self.get_id().decode()} has already been run on resource {resource.get_id().hex()}"
        )


class ComponentMissingDependencyError(RuntimeError):
    def __init__(
        self,
        component: ComponentInterface,
        dependency: ComponentExternalTool,
    ):
        if dependency.apt_package:
            apt_install_str = f"\n\tapt installation: apt install {dependency.apt_package}"
        else:
            apt_install_str = ""
        if dependency.brew_package:
            brew_install_str = f"\n\tbrew installation: brew install {dependency.brew_package}"
        else:
            brew_install_str = ""

        super().__init__(
            f"Missing {dependency.tool} tool needed for {type(component).__name__}!"
            f"{apt_install_str}"
            f"{brew_install_str}"
            f"\n\tSee {dependency.tool_homepage} for more info and installation help."
            f"\n\tAlternatively, OFRAK can ignore this component (and any others with missing "
            f"dependencies) so that they will never be run: OFRAK(..., exclude_components_missing_dependencies=True)"
        )

        self.component = component
        self.dependency = dependency


class ComponentSubprocessError(RuntimeError):
    def __init__(self, error: CalledProcessError):
        errstring = (
            f"Command '{error.cmd}' returned non-zero exit status {error.returncode}.\n"
            f"Stderr: {error.stderr}.\n"
            f"Stdout: {error.stdout}."
        )
        super().__init__(errstring)
        self.cmd = error.cmd
        self.cmd_retcode = error.returncode
        self.cmd_stdout = error.stdout
        self.cmd_stderr = error.stderr



import reedsolo as rs
from ofrak.core.ecc.abstract import EccAlgorithm, EccError


class ReedSolomon(EccAlgorithm):
    def __init__(
        self,
        nsym: int = 10,
        nsize: int = 255,
        fcr: int = 0,
        prim: int = 0x11D,
        generator: int = 2,
        c_exp: int = 8,
        single_gen: bool = True,
    ):
        self.RSC = rs.RSCodec(
            nsym=nsym,
            nsize=nsize,
            fcr=fcr,
            prim=prim,
            generator=generator,
            c_exp=c_exp,
            single_gen=single_gen,
        )

    def encode(self, payload: bytes) -> bytes:
        return self.RSC.encode(
            data=payload,
        )[len(payload) :]

    def decode(self, payload: bytes) -> bytes:
        try:
            return self.RSC.decode(data=payload)[0]
        except rs.ReedSolomonError:
            raise EccError()



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, Tuple, Optional

from ofrak.component.modifier import Modifier
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.resource import Resource
from ofrak_type.error import NotFoundError
from ofrak_type.range import Range


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class CommentsAttributes(ResourceAttributes):
    """
    User-defined comments, each comment associated with an optional range.
    """

    comments: Dict[Optional[Range], str]


@dataclass
class AddCommentModifierConfig(ComponentConfig):
    comment: Tuple[Optional[Range], str]


class AddCommentModifier(Modifier[AddCommentModifierConfig]):
    """
    Modifier to add a single comment to a resource.
    """

    targets = ()

    async def modify(self, resource: Resource, config: AddCommentModifierConfig) -> None:
        # Verify that the given range is valid for the given resource.
        config_range = config.comment[0]
        if config_range is not None:
            if config_range.start < 0 or config_range.end > len(await resource.get_data()):
                raise ValueError(
                    f"Range {config_range} is outside the bounds of "
                    f"resource {resource.get_id().hex()}"
                )

        try:
            # deepcopy the existing comments, otherwise they would be modified in place
            # and OFRAK would then compare the new attributes with the existing ones and find
            # they are the same, and report that the resource wasn't modified.
            comments = deepcopy(resource.get_attributes(CommentsAttributes).comments)
        except NotFoundError:
            comments = {}
        comments[config.comment[0]] = config.comment[1]
        resource.add_attributes(CommentsAttributes(comments=comments))


@dataclass
class DeleteCommentModifierConfig(ComponentConfig):
    comment_range: Optional[Range]


class DeleteCommentModifier(Modifier[DeleteCommentModifierConfig]):
    """
    Modifier to delete a comment from a resource.
    """

    targets = ()

    async def modify(self, resource: Resource, config: DeleteCommentModifierConfig) -> None:
        """
        Delete the comment associated with the given range.

        :raises NotFoundError: if the comment range is not associated with a comment.
        """
        try:
            comments = deepcopy(resource.get_attributes(CommentsAttributes).comments)
        except NotFoundError:
            comments = {}
        try:
            del comments[config.comment_range]
        except KeyError:
            raise NotFoundError(
                f"Comment range {config.comment_range} not found in "
                f"resource {resource.get_id().hex()}"
            )
        resource.add_attributes(CommentsAttributes(comments=comments))



import setuptools
import pkg_resources
from setuptools.command.egg_info import egg_info


class egg_info_ex(egg_info):
    """Includes license file into `.egg-info` folder."""

    def run(self):
        # don't duplicate license into `.egg-info` when building a distribution
        if not self.distribution.have_run.get("install", True):
            # `install` command is in progress, copy license
            self.mkpath(self.egg_info)
            self.copy_file("LICENSE", self.egg_info)

        egg_info.run(self)


with open("README.md") as f:
    long_description = f.read()


entropy_so = setuptools.Extension(
    "ofrak.core.entropy.entropy_c",
    sources=["ofrak/core/entropy/entropy.c"],
    libraries=["m"],  # math library
    optional=True,  # If this fails the build, OFRAK will fall back to Python implementation
    extra_compile_args=["-O3"],
)


# Should be the same as in build_image.py
def read_requirements(requirements_path):
    with open(requirements_path) as requirements_handle:
        return [
            str(requirement)
            for requirement in pkg_resources.parse_requirements(requirements_handle)
        ]


setuptools.setup(
    name="ofrak",
    version="2.2.1",
    description="A binary analysis and modification platform",
    packages=setuptools.find_packages(exclude=["test_ofrak", "test_ofrak.*"]),
    package_data={
        "ofrak": ["py.typed"],
    },
    install_requires=[
        "ofrak_io~=1.0",
        "ofrak_type~=2.0",
        "ofrak_patch_maker~=3.0",
    ]
    + read_requirements("requirements.txt"),
    extras_require={
        "docs": read_requirements("requirements-docs.txt"),
        "test": read_requirements("requirements-test.txt"),
    },
    author="Red Balloon Security",
    author_email="ofrak@redballoonsecurity.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://ofrak.com/",
    download_url="https://github.com/redballoonsecurity/ofrak",
    project_urls={
        "Documentation": "https://ofrak.com/docs/",
        "Community License": "https://github.com/redballoonsecurity/ofrak/blob/master/LICENSE",
        "Commercial Licensing Information": "https://ofrak.com/license/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: Other/Proprietary License",
        "License :: Free To Use But Restricted",
        "License :: Free For Home Use",
        "Topic :: Security",
        "Typing :: Typed",
    ],
    python_requires=">=3.7",
    license="Proprietary",
    license_files=["LICENSE"],
    cmdclass={"egg_info": egg_info_ex},
    entry_points={
        "ofrak.packages": ["ofrak_pkg = ofrak"],
        "console_scripts": ["ofrak = ofrak.__main__:main"],
    },
    ext_modules=[entropy_so],
    include_package_data=True,
)



from collections import defaultdict
from typing import Any, List, Set, Tuple, cast, Callable
from typing import Dict, Optional

from sortedcontainers import SortedList
from typing_inspect import get_origin

from ofrak import ResourceTag
from ofrak.model.resource_model import ResourceIndexedAttribute
from ofrak.service.resource_service import (
    ResourceNode,
    ResourceAttributeIndex,
    AttributeIndexDict,
    T,
)
from ofrak.service.resource_service import ResourceService
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class ResourceServiceSerializer(SerializerInterface):
    """
    Serialize and deserialize `ResourceService` into `PJSONType`.

    The challenging thing to serialize and deserialize is the `ResourceNode` store. This class
    complements the `ResourceNodeSerializer`, which doesn't fully serialize the parents and children of a
    node. So during deserialization, the major role of this class is to finalize the deserialization
    of the `ResourceNode` objects, by recovering their parents and children and updating the `ResourceNode`s.

    Also, _attribute_indexes is actually an AttributeIndexDict(ResourceAttributeIndex) (a kind of defaultdict),
    and _tag_indexes is a defaultdict(set). They're both serialized as dicts, but deserialized as the correct
    defaultdicts.
    """

    targets = (ResourceService,)

    resource_service_annotations = {
        "_resource_store": Dict[bytes, ResourceNode],
        "_resource_by_data_id_store": Dict[bytes, ResourceNode],
        "_attribute_indexes": Dict[ResourceIndexedAttribute[T], ResourceAttributeIndex[T]],
    }

    def obj_to_pjson(self, obj: ResourceService, _type_hint: Any) -> Dict[str, PJSONType]:
        resource_service_pjson = {
            attr_name: self._service.to_pjson(getattr(obj, attr_name), type_hint)
            for attr_name, type_hint in self.resource_service_annotations.items()
        }
        resource_service_pjson["_tag_indexes"] = self._tag_index_to_pjson(obj._tag_indexes)
        resource_service_pjson["_root_resources"] = self._root_resources_to_pjson(
            obj._root_resources
        )
        return resource_service_pjson

    def pjson_to_obj(self, pjson_obj: Dict[str, PJSONType], _type_hint: Any) -> ResourceService:
        deserialized_attrs = {
            attr_name: self._service.from_pjson(pjson_obj[attr_name], type_hint)
            for attr_name, type_hint in self.resource_service_annotations.items()
        }
        deserialized_attrs["_tag_indexes"] = self._tag_index_from_pjson(pjson_obj["_tag_indexes"])
        deserialized_attrs["_root_resources"] = self._root_resources_from_pjson(
            pjson_obj["_root_resources"]
        )

        resource_service: ResourceService = ResourceService.__new__(ResourceService)
        for attr_name, attr in deserialized_attrs.items():
            if attr_name not in ("_attribute_indexes", "_tag_indexes"):
                setattr(resource_service, attr_name, attr)

        for resource_node in resource_service._resource_store.values():
            # Update `parent`
            parent_id: Optional[bytes] = resource_node.model.parent_id
            if parent_id is None:
                parent = None
            else:
                parent = resource_service._resource_store[parent_id]
            resource_node.parent = parent
            # Update `_children`
            resource_node._children = {
                resource_service._resource_store[child_id]: None
                for child_id in getattr(resource_node, "_pjson_children_ids")
            }
            delattr(resource_node, "_pjson_children_ids")

        # convert ID shorthand in _root_resources to actual nodes
        setattr(
            resource_service,
            "_root_resources",
            {
                root_id: resource_service._resource_store[root_id]
                for root_id in deserialized_attrs["_root_resources"]
            },
        )

        # convert ID shorthand in attribute indexes to actual nodes
        for attribute_index in deserialized_attrs["_attribute_indexes"].values():
            attribute_index.index = SortedList(
                [
                    (val, resource_service._resource_store[node_id])
                    for val, node_id in attribute_index.index
                ]
            )
        # _attribute_indexes is actually an AttributeIndexDict(ResourceAttributeIndex)
        setattr(
            resource_service,
            "_attribute_indexes",
            AttributeIndexDict(
                cast(Callable, ResourceAttributeIndex), deserialized_attrs["_attribute_indexes"]
            ),
        )

        # convert ID shorthand in tag indexes to actual nodes
        finished_tag_indexes = {
            tag: {resource_service._resource_store[node_id] for node_id in node_ids}
            for tag, node_ids in deserialized_attrs["_tag_indexes"].items()
        }
        # _tag_indexes is actually a defaultdict(set)
        setattr(resource_service, "_tag_indexes", defaultdict(set, finished_tag_indexes))

        return resource_service

    def _tag_index_to_pjson(self, tag_index: Dict[ResourceTag, Set[ResourceNode]]):
        simplified_index = {
            tag: {node.model.id for node in nodes} for tag, nodes in tag_index.items()
        }
        return self._service.to_pjson(simplified_index, Dict[ResourceTag, Set[bytes]])

    def _tag_index_from_pjson(self, tag_index_pjson: PJSONType):
        return self._service.from_pjson(tag_index_pjson, Dict[ResourceTag, Set[bytes]])

    def _root_resources_to_pjson(self, root_resources: Dict[bytes, ResourceNode]):
        return self._service.to_pjson(root_resources.keys(), Set[bytes])

    def _root_resources_from_pjson(self, root_resources_pjson: PJSONType) -> Set[bytes]:
        return self._service.from_pjson(root_resources_pjson, Set[bytes])


class ResourceNodeSerializer(SerializerInterface):
    """
    Serialize and deserialize `ResourceNode` into `PJSONType`.

    This requires a custom serializer because of the infinite recursion the default serializer
    would encounter, with a ResourceNode storing instances of both its parent and children.

    Implementation:
    - the parent isn't serialized;
    - children are serialized as the IDs of their ResourceModel instead of the children themselves.

    The deserialized `ResourceNode` will be temporary and will need to be updated by the `ResourceService` deserializer.
    Its fields `parent` and `_children` aren't set, instead:
     - the parent can be retrieved from the `parent_id` of the ResourceModel;
    - children can be found from their IDs stored in a temporary field `_pjson_children_ids`.
    Recovering these attributes is the role of the `ResourceServiceSerializer`.
    """

    targets = (ResourceNode,)

    serialized_annotations = {
        attr_name: attr_type
        for attr_name, attr_type in ResourceNode.__annotations__.items()
        if attr_name not in ("parent", "_children")
    }

    def obj_to_pjson(self, obj: ResourceNode, _type_hint: Any) -> Dict[str, PJSONType]:
        result = {}
        for attr_name, type_hint in self.serialized_annotations.items():
            result[attr_name] = self._service.to_pjson(getattr(obj, attr_name), type_hint)
        children_ids = [child.model.id for child in obj._children.keys()]
        result["_pjson_children_ids"] = self._service.to_pjson(children_ids, List[bytes])
        return result

    def pjson_to_obj(self, pjson_obj: Dict[str, PJSONType], _type_hint: Any) -> ResourceNode:
        deserialized_attrs = {
            attr_name: self._service.from_pjson(pjson_obj[attr_name], type_hint)
            for attr_name, type_hint in self.serialized_annotations.items()
        }
        deserialized_attrs.update({"parent": None, "_children": []})
        resource_node = ResourceNode.__new__(ResourceNode)
        for attr_name, attr in deserialized_attrs.items():
            setattr(resource_node, attr_name, attr)
        setattr(
            resource_node,
            "_pjson_children_ids",
            self._service.from_pjson(pjson_obj["_pjson_children_ids"], List[bytes]),
        )
        return resource_node


def _is_generic_resource_attribute_index(type_hint):
    return get_origin(type_hint) == ResourceAttributeIndex


class ResourceAttributeIndexSerializer(SerializerInterface):
    """
    Serialize and deserialize `ResourceAttributeIndex[T]` into `PJSONType`.

    The default serializer works for `ResourceAttributeIndex`, we just need to handle `ResourceAttributeIndex[T]`.
    This is done by ignoring the type argument and using `ResourceAttributeIndex` as type hint instead.
    """

    targets = (_is_generic_resource_attribute_index, ResourceAttributeIndex)

    def obj_to_pjson(self, obj: ResourceAttributeIndex, type_hint: Any) -> PJSONType:
        simplified_index = [(val, node.model.id) for val, node in obj.index]
        result = {
            "_attribute": self._service.to_pjson(obj._attribute, ResourceIndexedAttribute),
            "index": self._service.to_pjson(simplified_index, List[Tuple[Any, bytes]]),
            "values_by_node_id": self._service.to_pjson(obj.values_by_node_id, Dict[bytes, Any]),
        }
        return result

    def pjson_to_obj(self, pjson_obj: PJSONType, type_hint: Any) -> ResourceAttributeIndex:
        if not isinstance(pjson_obj, dict):
            raise ValueError(f"Expected to deserialize a dict, got {type(pjson_obj)}")
        deserialized_attrs = {
            "_attribute": self._service.from_pjson(
                pjson_obj["_attribute"], ResourceIndexedAttribute
            ),
            "index": self._service.from_pjson(pjson_obj["index"], List[Tuple[Any, bytes]]),
            "values_by_node_id": self._service.from_pjson(
                pjson_obj["values_by_node_id"], Dict[bytes, Any]
            ),
        }

        reconstructed_index = ResourceAttributeIndex(deserialized_attrs["_attribute"])
        reconstructed_index.index = SortedList(deserialized_attrs["index"])
        reconstructed_index.values_by_node_id = deserialized_attrs["values_by_node_id"]

        return reconstructed_index  # Only partially reconstructed



import logging
import zlib
from dataclasses import dataclass

from ofrak.component.analyzer import Analyzer
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)


@dataclass
class ZlibData(GenericBinary):
    compression_level: int

    async def get_child(self) -> GenericBinary:
        return await self.resource.get_only_child_as_view(GenericBinary)


class ZlibCompressionLevelAnalyzer(Analyzer[None, ZlibData]):
    id = b"ZlibCompressionLevelAnalyzer"
    targets = (ZlibData,)
    outputs = (ZlibData,)

    async def analyze(self, resource: Resource, config=None) -> ZlibData:
        zlib_data = await resource.get_data(Range(0, 2))
        flevel = zlib_data[-1]
        if flevel == 0x01:
            compression_level = 1
        elif flevel == 0x5E:
            compression_level = 2
        elif flevel == 0x9C:
            compression_level = 6
        elif flevel == 0xDA:
            compression_level = 7
        else:
            compression_level = 6
        return ZlibData(compression_level)


class ZlibUnpacker(Unpacker[None]):
    """
    Decompress a blob of zlib data
    """

    id = b"ZlibUnpacker"
    targets = (ZlibData,)
    children = (GenericBinary,)

    async def unpack(self, resource: Resource, config=None):
        zlib_data = await resource.get_data()
        zlib_uncompressed_data = zlib.decompress(zlib_data)
        await resource.create_child(
            tags=(GenericBinary,),
            data=zlib_uncompressed_data,
        )


class ZlibPacker(Packer[None]):
    """
    Pack a binary with zlib
    """

    targets = (ZlibData,)

    async def pack(self, resource: Resource, config=None):
        zlib_view = await resource.view_as(ZlibData)
        compression_level = zlib_view.compression_level
        zlib_child = await zlib_view.get_child()
        zlib_data = await zlib_child.resource.get_data()
        zlib_compressed = zlib.compress(zlib_data, compression_level)

        original_zlib_size = await zlib_view.resource.get_data_length()
        resource.queue_patch(Range(0, original_zlib_size), zlib_compressed)


MagicMimeIdentifier.register(ZlibData, "application/zlib")
MagicDescriptionIdentifier.register(ZlibData, lambda s: s.startswith("zlib compressed data"))



from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Optional

from ofrak.core.program import Program
from ofrak.core.program_section import NamedProgramSection
from ofrak.model.resource_model import index
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import (
    ResourceAttributeValueFilter,
    ResourceFilter,
)
from ofrak.core.magic import MagicDescriptionIdentifier
from ofrak_type.error import NotFoundError


@dataclass
class PeMsDosHeader(ResourceView):
    """PE MS-DOS header"""

    e_magic: int
    e_cblp: int
    e_cp: int
    e_crlc: int
    e_cparhdr: int
    e_minalloc: int
    e_maxalloc: int
    e_ss: int
    e_sp: int
    e_csum: int
    e_ip: int
    e_cs: int
    e_lfarlc: int
    e_ovno: int
    e_res: bytes
    e_oemid: int
    e_oeminfo: int
    e_res2: bytes
    e_lfanew: int


@dataclass
class PeFileHeader(ResourceView):
    """PE file header, a.k.a. COFF file header"""

    machine: int
    number_of_sections: int
    time_date_stamp: int
    pointer_to_symbol_table: int
    number_of_symbols: int
    size_of_optional_header: int
    characteristics: int


class PeOptionalHeaderMagic(Enum):
    """
    Magic values for the PE optional header
    """

    IMAGE_NT_OPTIONAL_HDR32_MAGIC = 0x10B
    IMAGE_NT_OPTIONAL_HDR64_MAGIC = 0x20B
    IMAGE_ROM_OPTIONAL_HDR_MAGIC = 0x107


@dataclass
class PeOptionalHeader(ResourceView):
    """
    Base of PE optional header. Includes NT-specific attributes.
    This header class is valid for 64 Bit non-Windows PE files,
    however our pe_file package populates the additional Windows
    fields so this is for all intents and purposes an abstract class.
    More Info: https://learn.microsoft.com/en-us/windows/win32/debug/pe-format#optional-header-windows-specific-fields-image-only

    Required for image files; object files don't have it.
    """

    magic: int
    major_linker_version: int
    minor_linker_version: int
    size_of_code: int
    size_of_initialized_data: int
    size_of_uninitialized_data: int
    address_of_entry_point: int
    base_of_code: int


@dataclass
class PeWinOptionalHeader(PeOptionalHeader):
    """PE Optional Header with all fields for Windows.
    These fields are not necessary, yet will be parsed by the pefile package regardless.
    """

    image_base: int
    section_alignment: int
    file_alignment: int
    major_operating_system_version: int
    minor_operating_system_version: int
    major_image_version: int
    minor_image_version: int
    major_subsystem_version: int
    minor_subsystem_version: int
    size_of_image: int
    size_of_headers: int
    checksum: int
    subsystem: int
    dll_characteristics: int
    size_of_stack_reserve: int
    size_of_stack_commit: int
    size_of_heap_reserve: int
    size_of_heap_commit: int
    loader_flags: int
    number_of_rva_and_sizes: int


@dataclass
class Pe64POptionalHeader(PeWinOptionalHeader):
    """
    64 bit PE optional header without the base_of_data field.
    """


@dataclass
class Pe32OptionalHeader(PeWinOptionalHeader):
    """
    32 bit PE optional header with the base_of_data field.
    """

    base_of_data: int


@dataclass
class PeDataDirectory(ResourceView):
    """PE data directory (image only)"""

    name: str
    virtual_address: int
    size: int


@dataclass
class PeSectionStructure(ResourceView):
    """Superclass for both the section headers and the sections themselves, linking them via the section index."""

    # Index of the section in the section table. Used to associate a section header with its section.
    section_index: int

    @index
    def SectionIndex(self) -> int:
        return self.section_index


@dataclass
class PeSection(PeSectionStructure, NamedProgramSection):
    """PE section"""

    async def get_header(self) -> "PeSectionHeader":
        return await self.resource.get_only_sibling_as_view(
            PeSectionHeader,
            ResourceFilter(
                tags=(PeSectionHeader,),
                attribute_filters=[
                    ResourceAttributeValueFilter(
                        PeSectionStructure.SectionIndex, self.section_index
                    )
                ],
            ),
        )


class PeSectionFlag(Enum):
    """
    Flags making up ``PeSectionHeader.m_characteristics``.

    Refer to https://docs.microsoft.com/en-us/windows/win32/debug/pe-format#section-flags
    for documentation.
    """

    IMAGE_SCN_TYPE_NO_PAD = 0x00000008
    IMAGE_SCN_CNT_CODE = 0x00000020
    IMAGE_SCN_CNT_INITIALIZED_DATA = 0x00000040
    IMAGE_SCN_CNT_UNINITIALIZED_DATA = 0x00000080
    IMAGE_SCN_LNK_OTHER = 0x00000100
    IMAGE_SCN_LNK_INFO = 0x00000200
    IMAGE_SCN_LNK_REMOVE = 0x00000800
    IMAGE_SCN_LNK_COMDAT = 0x00001000
    IMAGE_SCN_GPREL = 0x00008000
    IMAGE_SCN_MEM_PURGEABLE = 0x00020000
    IMAGE_SCN_MEM_16BIT = 0x00020000
    IMAGE_SCN_MEM_LOCKED = 0x00040000
    IMAGE_SCN_MEM_PRELOAD = 0x00080000
    IMAGE_SCN_ALIGN_1BYTES = 0x00100000
    IMAGE_SCN_ALIGN_2BYTES = 0x00200000
    IMAGE_SCN_ALIGN_4BYTES = 0x00300000
    IMAGE_SCN_ALIGN_8BYTES = 0x00400000
    IMAGE_SCN_ALIGN_16BYTES = 0x00500000
    IMAGE_SCN_ALIGN_32BYTES = 0x00600000
    IMAGE_SCN_ALIGN_64BYTES = 0x00700000
    IMAGE_SCN_ALIGN_128BYTES = 0x00800000
    IMAGE_SCN_ALIGN_256BYTES = 0x00900000
    IMAGE_SCN_ALIGN_512BYTES = 0x00A00000
    IMAGE_SCN_ALIGN_1024BYTES = 0x00B00000
    IMAGE_SCN_ALIGN_2048BYTES = 0x00C00000
    IMAGE_SCN_ALIGN_4096BYTES = 0x00D00000
    IMAGE_SCN_ALIGN_8192BYTES = 0x00E00000
    IMAGE_SCN_LNK_NRELOC_OVFL = 0x01000000
    IMAGE_SCN_MEM_DISCARDABLE = 0x02000000
    IMAGE_SCN_MEM_NOT_CACHED = 0x04000000
    IMAGE_SCN_MEM_NOT_PAGED = 0x08000000
    IMAGE_SCN_MEM_SHARED = 0x10000000
    IMAGE_SCN_MEM_EXECUTE = 0x20000000
    IMAGE_SCN_MEM_READ = 0x40000000
    IMAGE_SCN_MEM_WRITE = 0x80000000


@dataclass
class PeSectionHeader(PeSectionStructure):
    """PE section header"""

    m_name: bytes
    m_virtual_size: int
    m_virtual_address: int
    m_size_of_raw_data: int
    m_pointer_to_raw_data: int
    m_pointer_to_relocations: int
    m_pointer_to_linenumbers: int
    m_number_of_relocations: int
    m_number_of_linenumbers: int
    m_characteristics: int

    @property
    def name(self) -> str:
        """The section name as a string, e.g. ".text" for b".text\x00\x00\x00"."""
        return self.m_name.rstrip(b"\x00").decode("ascii")

    def has_flag(self, flag: PeSectionFlag) -> bool:
        return self.m_characteristics & flag.value != 0

    def get_flags(self) -> Iterable[PeSectionFlag]:
        for flag in PeSectionFlag:
            if self.has_flag(flag):
                yield flag

    async def get_body(self) -> PeSection:
        """Get the PeSection associated with this section header."""
        return await self.resource.get_only_sibling_as_view(
            PeSection,
            ResourceFilter(
                tags=(PeSection,),
                attribute_filters=[
                    ResourceAttributeValueFilter(
                        PeSectionStructure.SectionIndex, self.section_index
                    )
                ],
            ),
        )


@dataclass
class Pe(Program):
    """PE file"""

    async def get_sections(self) -> Iterable[PeSection]:
        """Return the children ``PeSection``s."""
        return await self.resource.get_children_as_view(
            PeSection,
            ResourceFilter(
                tags=(PeSection,),
            ),
        )

    async def get_section_by_name(self, name: str) -> PeSection:
        # Force analyzing the name of all sections
        await self.get_sections()

        return await self.resource.get_only_child_as_view(
            PeSection,
            ResourceFilter(
                tags=(PeSection,),
                attribute_filters=(ResourceAttributeValueFilter(PeSection.SectionName, name),),
            ),
        )

    async def get_optional_header(self) -> Optional[PeOptionalHeader]:
        """Return the optional header of the PE file, or None if there isn't one."""
        try:
            return await self.resource.get_only_child_as_view(
                PeOptionalHeader,
                ResourceFilter(
                    tags=(PeOptionalHeader,),
                ),
            )
        except NotFoundError:
            return None


MagicDescriptionIdentifier.register(Pe, lambda s: s.startswith("PE32"))



from abc import ABCMeta, abstractmethod, ABC
from types import ModuleType
from typing import List, Type, Optional, TypeVar, Set

from dataclasses import dataclass

from ofrak.component.interface import ComponentInterface

CI = TypeVar("CI", bound="ComponentInterface")


# MyPy shows a type error on the next line, root cause unknown:
# error: Only concrete class can be given where "Type[ComponentFilter]" is expected
@dataclass(frozen=True, init=False)  # type: ignore
class ComponentFilter(ABC):
    @abstractmethod
    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        """
        Filter out components according to the rules this component enforces.

        :param components: Components to filter

        :return: All components which this filter allows
        """
        raise NotImplementedError()


class ComponentLocatorInterface(metaclass=ABCMeta):
    @abstractmethod
    def add_components(
        self,
        components: List[ComponentInterface],
        module_priority: Optional[List[ModuleType]] = None,
    ):
        pass

    @abstractmethod
    def get_by_id(self, component_id: bytes) -> ComponentInterface:
        pass

    @abstractmethod
    def get_by_type(self, component_type: Type[CI]) -> CI:
        pass

    @abstractmethod
    def get_components_matching_filter(
        self, component_filter: ComponentFilter
    ) -> Set[ComponentInterface]:
        """
        Get all components matching the given filter.
        """



from typing import Any, List, Set, FrozenSet

from typing_inspect import get_origin, get_args

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class SetSerializer(SerializerInterface):
    """
    Serialize and deserialize `Set[X]` into `PJSONType`.

    Implementation: sets are serialized as lists.
    """

    targets = (lambda type_hint: get_origin(type_hint) == set,)

    def obj_to_pjson(self, obj: Set[Any], type_hint: Any) -> List[PJSONType]:
        args = get_args(type_hint)
        return [self._service.to_pjson(item, args[0]) for item in obj]

    def pjson_to_obj(self, pjson_obj: List[PJSONType], type_hint: Any) -> Set[Any]:
        args = get_args(type_hint)
        return {self._service.from_pjson(item, args[0]) for item in pjson_obj}


class FrozenSetSerializer(SerializerInterface):
    """
    Serialize and deserialize `FrozenSet[X]` into `PJSONType`.

    Implementation: frozensets are serialized as lists.
    """

    targets = (lambda type_hint: get_origin(type_hint) == frozenset,)

    def obj_to_pjson(self, obj: FrozenSet[Any], type_hint: Any) -> List[PJSONType]:
        args = get_args(type_hint)
        return [self._service.to_pjson(item, args[0]) for item in obj]

    def pjson_to_obj(self, pjson_obj: List[PJSONType], type_hint: Any) -> FrozenSet[Any]:
        args = get_args(type_hint)
        return frozenset(self._service.from_pjson(item, args[0]) for item in pjson_obj)



from dataclasses import dataclass

from ofrak.model.resource_model import ResourceAttributes

from ofrak_type.architecture import ArchInfo


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class ProgramAttributes(ResourceAttributes, ArchInfo):
    """
    Analyzer output containing architecture attributes of a program.

    """



from typing import Any

from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class BytesSerializer(SerializerInterface):
    """
    Serialize and deserialize bytes into `PJSONType`.

    Implementation: hex string.
    """

    targets = (bytes,)

    def obj_to_pjson(self, obj: bytes, _type_hint: Any) -> str:
        return obj.hex()

    def pjson_to_obj(self, pjson_obj: str, _type_hint: Any) -> bytes:
        return bytes.fromhex(pjson_obj)



from typing import Any

from abc import ABC, abstractmethod

from ofrak.service.serialization.pjson_types import PJSONType


class SerializationServiceInterface(ABC):
    """
    Interface for a serialization service.

    Note that the interface is currently tightly coupled with the PJSON implementation. This is because other parts
    of the code expect and process the PJSON format, so the serialization format can't be trivially replaced with
    another one at the moment.
    """

    @abstractmethod
    def to_pjson(self, obj: Any, type_hint: Any) -> PJSONType:
        raise NotImplementedError()

    @abstractmethod
    def from_pjson(self, pjson_obj: PJSONType, type_hint: Any) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def to_json(self, obj: Any, type_hint: Any) -> str:
        raise NotImplementedError()

    @abstractmethod
    def from_json(self, json_obj: str, type_hint: Any) -> Any:
        raise NotImplementedError()



import asyncio
from argparse import RawDescriptionHelpFormatter, Namespace
from types import ModuleType
from typing import Iterable, Set, Dict

from ofrak.cli.ofrak_cli import OfrakCommand, OFRAKEnvironment
from ofrak.model.component_model import ComponentExternalTool


class DepsCommand(OfrakCommand):
    def create_parser(self, ofrak_subparsers):
        deps_parser = ofrak_subparsers.add_parser(
            "deps",
            help="Show and check the external (non-Python) dependencies of OFRAK components. Can "
            "show the brew/apt install packages for dependencies, and filter by component or "
            "package.",
            description="Show/check the dependencies of OFRAK components.\n"
            "Examples:\n"
            "\tGet all dependencies of core ofrak:\n"
            "\t\tpython3 -m ofrak deps --package ofrak\n"
            "\tList all the apt packages needed for missing dependencies:\n"
            "\t\tpython3 -m ofrak deps --missing-only --packages-for apt",
            formatter_class=RawDescriptionHelpFormatter,
        )

        deps_parser.add_argument(
            "--package", action="append", help="Include dependencies of this package"
        )
        deps_parser.add_argument(
            "--component", action="append", help="Include dependencies of this component"
        )
        deps_parser.add_argument(
            "--missing-only",
            action="store_true",
            help="Only output information for missing dependencies",
        )
        mutex_group = deps_parser.add_mutually_exclusive_group()
        mutex_group.add_argument(
            "--packages-for",
            action="store",
            choices=("apt", "brew"),
            help="List only names of packages (known to the selected package manager) which "
            "provide required dependencies.",
            default=None,
        )
        mutex_group.add_argument(
            "--no-packages-for",
            action="store",
            choices=("apt", "brew"),
            help="List only dependencies which cannot be installed via the selected package "
            "manager.",
            default=None,
        )
        deps_parser.add_argument(
            "--no-check",
            "-n",
            action="store_true",
            help="Do not check that each dependency is present (ignored if --missing-only is also "
            "provided)",
        )
        return deps_parser

    def run(self, ofrak_env: OFRAKEnvironment, args: Namespace):
        dependencies = set()
        packages: Iterable[ModuleType]
        check_deps = args.missing_only or not args.no_check

        if args.package or args.component:
            packages = []
        else:
            packages = ofrak_env.packages.values()

        if args.package:
            packages = [ofrak_env.packages[pkg_name] for pkg_name in args.package]

        if args.component:
            components = [ofrak_env.components[component_name] for component_name in args.component]
        else:
            components = []

        for pkg in packages:
            components.extend(ofrak_env.components_by_package[pkg])

        if not components:
            return
        components_by_dep: Dict[ComponentExternalTool, Set[str]] = dict()
        for component in ofrak_env.components.values():
            for dep in component.external_dependencies:  # type: ignore
                if dep not in components_by_dep:
                    components_by_dep[dep] = set()
                components_by_dep[dep].add(component.__name__)

        for component in components:
            dep_list = ofrak_env.dependencies_by_component[component]
            for dep in dep_list:
                dependencies.add(dep)

        for dep in dependencies:
            if check_deps:
                is_installed = asyncio.run(dep.is_tool_installed())
            else:
                is_installed = None

            if args.missing_only and is_installed:
                continue

            if args.no_packages_for:
                pkg_manager = args.no_packages_for
            elif args.packages_for:
                pkg_manager = args.packages_for
            else:
                pkg_manager = None

            if pkg_manager == "apt":
                dep_pkg = dep.apt_package
            elif pkg_manager == "brew":
                dep_pkg = dep.brew_package
            else:
                dep_pkg = None

            if args.packages_for:
                if dep_pkg is not None:
                    print(dep_pkg)
                continue
            elif args.no_packages_for and dep_pkg is not None:
                continue

            dependency_info = f"{dep.tool}\n\t{dep.tool_homepage}\n\t[{', '.join(c for c in components_by_dep[dep])}]"
            if not args.no_check:
                dependency_info = f"[{' ' if not is_installed else '✓'}] " + dependency_info

            print(dependency_info)



import struct
from collections import defaultdict
from dataclasses import dataclass
from typing import Tuple, Dict

from ofrak.core.memory_region import MemoryRegion
from ofrak.model.resource_model import ResourceAttributes
from ofrak_type.range import Range


@dataclass
class DataWord(MemoryRegion):
    """
    Addressable data word. Size depends on target system, and is extracted by analysis backend.

    :ivar virtual_address: virtual address of the memory region containing the data word
    :ivar size: size of the memory region containing the data word
    :ivar format_string: `struct.unpack` format string for unpacking the data word into a value
    :ivar xrefs_to: cross references to the data word
    """

    format_string: str
    xrefs_to: Tuple[int, ...]

    async def get_value_unsigned(self) -> int:
        data = await self.resource.get_data()
        return struct.unpack(self.format_string.upper(), data)[0]

    async def get_value_signed(self) -> int:
        data = await self.resource.get_data()
        return struct.unpack(self.format_string.lower(), data)[0]


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class ReferencedStringsAttributes(ResourceAttributes):
    """
    Strings referenced by functions, represented as
    [ResourceAttributes][ofrak.model.resource_model.ResourceAttributes].

    :ivar referencing_functions: All functions which reference strings in the program
    :ivar referenced_strings: All strings which are referenced in the program
    :ivar references: Edges from referencing funcs to referenced strings
    """

    referencing_functions: Tuple[int, ...]
    referenced_strings: Tuple[Range, ...]
    references: Tuple[Tuple[int, int], ...]


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class ReferencedDataAttributes(ResourceAttributes):
    """
    Data cross-references in a program. That is, all addresses which are known to reference data
    and all data addresses which are known to be referenced.

    :ivar referencing_addresses: Addresses of all items which reference data in the program
    :ivar referenced_data: All data which are referenced in the program
    :ivar references: Pairs of indexes in `referencing_addresses` and `referenced_strings`
    representing edges from referencing funcs to referenced data
    """

    referencing_addresses: Tuple[int, ...]
    referenced_data: Tuple[int, ...]
    references: Tuple[Tuple[int, int], ...]

    def get_xrefs_to(self) -> Dict[int, Tuple[int, ...]]:
        """
        Create a dictionary version of the data ref information. The keys are data addresses, and
        the value for each of those is a tuple of all the addresses which reference that data
        address.

        :return: Dictionary mapping each data address to the addresses that reference it
        """
        data_vaddrs_to_referrees = defaultdict(list)
        for ref_from_idx, ref_to_idx in self.references:
            data_vaddrs_to_referrees[self.referenced_data[ref_to_idx]].append(
                self.referencing_addresses[ref_from_idx]
            )
        return {
            data_vaddr: tuple(addresses_refering_to_data_vaddr)
            for data_vaddr, addresses_refering_to_data_vaddr in data_vaddrs_to_referrees.items()
        }

    def get_xrefs_from(self):
        """
        Create a dictionary version of the data ref information. The keys are addresses, and
        the value for each of those is a tuple of all the data addresses are referenced from that
        address.

        :return: Dictionary mapping each address to the addresses of the data that it references
        """
        addresses_to_referenced_data = defaultdict(list)
        for ref_from_idx, ref_to_idx in self.references:
            addresses_to_referenced_data[self.referencing_addresses[ref_from_idx]].append(
                self.referenced_data[ref_to_idx]
            )
        return {
            vaddr: tuple(referenced_data_vaddrs)
            for vaddr, referenced_data_vaddrs in addresses_to_referenced_data.items()
        }



import asyncio
import tempfile
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier

from ofrak.model.component_model import CC, ComponentExternalTool
from ofrak_type.range import Range

LZOP = ComponentExternalTool(
    "lzop", "https://www.lzop.org/", "--help", apt_package="lzop", brew_package="lzop"
)


class LzoData(GenericBinary):
    """
    An lzo binary blob.
    """

    async def get_child(self) -> GenericBinary:
        return await self.resource.get_only_child_as_view(GenericBinary)


class LzoUnpacker(Unpacker[None]):
    """
    Unpack (decompress) an LZO file.
    """

    id = b"LzoUnpacker"
    targets = (LzoData,)
    children = (GenericBinary,)
    external_dependencies = (LZOP,)

    async def unpack(self, resource: Resource, config: CC) -> None:
        with tempfile.NamedTemporaryFile(suffix=".lzo") as compressed_file:
            compressed_file.write(await resource.get_data())
            compressed_file.flush()

            cmd = [
                "lzop",
                "-d",
                "-f",
                "-c",
                compressed_file.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await proc.communicate()
            if proc.returncode:
                raise CalledProcessError(returncode=proc.returncode, cmd=cmd)

            await resource.create_child(tags=(GenericBinary,), data=stdout)


class LzoPacker(Packer[None]):
    """
    Pack data into a compressed LZO file.
    """

    targets = (LzoData,)
    external_dependencies = (LZOP,)

    async def pack(self, resource: Resource, config=None):
        lzo_view = await resource.view_as(LzoData)
        child_file = await lzo_view.get_child()
        uncompressed_data = await child_file.resource.get_data()

        with tempfile.NamedTemporaryFile(suffix=".lzo") as uncompressed_file:
            uncompressed_file.write(uncompressed_data)
            uncompressed_file.flush()

            cmd = [
                "lzop",
                "-f",
                "-c",
                uncompressed_file.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await proc.communicate()
            if proc.returncode:
                raise CalledProcessError(returncode=proc.returncode, cmd=cmd)

            compressed_data = stdout
            original_size = await lzo_view.resource.get_data_length()
            resource.queue_patch(Range(0, original_size), compressed_data)


MagicMimeIdentifier.register(LzoData, "application/x-lzop")
MagicDescriptionIdentifier.register(LzoData, lambda s: s.lower().startswith("lzop compressed data"))



from typing import Any, Tuple

from ofrak.service.serialization.serializers.serializer_i import SerializerInterface
from ofrak_type.range import Range


class RangeSerializer(SerializerInterface):
    """
    Serialize and deserialize `Range` into `PJSONType`.

    Implementation: a `Range` is serialized as a 2-tuple (start, end).

    The default class instance serializer would work, however the space overhead would be
    significant and `Range` is used frequently enough in OFRAK to justify this optimization.
    """

    targets = (Range,)

    def obj_to_pjson(self, obj: Range, _type_hint: Any) -> Tuple[int, int]:
        return (obj.start, obj.end)

    def pjson_to_obj(self, pjson_obj: Tuple[int, int], _type_hint: Any) -> Range:
        return Range(pjson_obj[0], pjson_obj[1])



import logging

import pytest

from ofrak.service.id_service_sequential import SequentialIDService
from ofrak import OFRAK

from synthol.injector import DependencyInjector


@pytest.fixture
def ofrak_injector():
    return DependencyInjector()


@pytest.fixture
def ofrak_id_service():
    return SequentialIDService()


@pytest.fixture
def ofrak(ofrak_injector, ofrak_id_service):
    ofrak = OFRAK(logging.INFO)
    ofrak.injector = ofrak_injector
    ofrak.set_id_service(ofrak_id_service)

    return ofrak


@pytest.fixture
async def ofrak_context(ofrak):
    context = await ofrak.create_ofrak_context()
    yield context
    await context.shutdown_context()
